///////////////////////////////////////////////////////////////////////////
// Filename:    Do178Builder.h
// Purpose:     Header for the Do178Builder project.
// Mods:        09/23/01 RSB    Began.  This file was initially created
//                              by wxDesigner, and is modified by wxDesigner
//                              every time the user interface is tweaked.
//                              However, a lot of other stuff has also been
//                              added to it manually, since all of the 
//                              Do178Builder files (and not just the ones
//                              related to wxWindows) include this file.
//                              The program comments in this file beginning
//                              "WDR:" are used by wxDesigner, and should
//                              not be modified.
//              10/21/01 RSB    Added control of output directory.
//              02/18/02 RSB    Added OnSize to the XmlEditorDialogClass.
//              04/26/04 RSB    Began allowing for file-locking.
//                              The method uses Alberto Bertogli's 
//                              Open Lock Daemon ("old"),
//                              http://users.auriga.wearlab.de/~alb/old/.
//          	07/04/06 RSB    Now builds with wxGTK-2.6.x and gcc 4.x.
//      	03/23/08 RSB    Added support for new toolbar buttons
//              		(AddCustom, AddSubCustom, InsertHyperlink,
//              		InsertFootnote, InsertMdash).
//		01/28/09 RSB	Added SFIX() macro.
//		01/29/09 RSB	Added the char_str macro, so that we can
//				adjust properly to the wxWidgets version
//				and Unicode vs. non-Unicode.

/*
  Copyright 2001,2002,2004,2008-2009 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/////////////////////////////////////////////////////////////////////////////

#ifndef __Do178Builder_H__
#define __Do178Builder_H__

#ifdef __GNUG__
#pragma interface "Do178Builder.cpp"
#endif

// Include wxWindows' headers

#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif

#include "Do178Builder_wdr.h"

// Some stuff for locking databases, based on Alberto Bertogli's 
// Open Lock Daemon ("old").  Unfortunately, "old" is a Linux-only thing
// right now, so we fake up some Win32 stuff to just transparently 
// ignore the whole issue of locking.
#ifdef WIN32
#if 0
#define old_connect(x,y) 3  // An arbitrary number not -1.
#define old_lock(x,y) 1     // "Success".
#define old_unlock(x,y) 1   // "Success".
#define old_trylock(x,y) 1  // "Success".
#else // 0
#include "old-0.16/lib/old.h"
#endif // 0
#else // WIN32
#include <old.h>
#endif // WIN32

// This is my equivalent of wxT(), but for variable C-strings rather
// than constant strings.
#define SFIX(S) wxString ((const char *) S, wxConvLocal)
// String size used for sucking C-strings out of wxStrings.
#define SOD_SIZE 1025

// In Unicode builds, it's necessary to use the wxString char_str()
// method to appropriately translate to normal 8-bit C-strings.
// However, this method appears only in wxWidgets 2.8.4 or later.
// The method c_str() can be used in place, but it only works in
// non-Unicode builds.
#if wxCHECK_VERSION(2, 8, 4)
#define char_str char_str
#else
#ifdef UNICODE
#error Cannot use a Unicode build of wxWidgets unless it is v2.8.4 or later.
#else
#define char_str c_str
#endif
#endif

// Helpful constants for the criticality levels and control categories.
#define DO178A_LEVEL_1 0
#define DO178A_LEVEL_2 1
#define DO178A_LEVEL_3 2
#define DO178B_LEVEL_A 3
#define DO178B_LEVEL_B 4
#define DO178B_LEVEL_C 5
#define DO178B_LEVEL_D 6
#define DO178B_LEVEL_E 7
#define NUM_CRITICALITY_LEVELS 8
#define SATISFIED_DISCRETION 0x00
#define SATISFIED_YES 0x01
#define SATISFIED_INDEPENDENCE 0x02
#define SATISFIED_MASK 0x0f
#define CC_DISCRETION 0x00
#define CC_2 0x10
#define CC_1 0x20
#define CC_MASK 0xf0

// Other constants, for the document types:
#define PSAC 0
#define SDP 1
#define SVP 2
#define SCMP 3
#define SQAP 4
#define SRS 5
#define SDS 6
#define SCS 7
#define SRD 8
#define DD 9
#define SDD 9
#define SVCP 10
#define SECI 11
#define SCI 12
#define SAS 13
#define NUM_DOCUMENTS 14
#define DOCUMENT_DISCRETION 0
#define DOCUMENT_YES 1

//----------------------------------------------------------------------------

// This is the data structure for a single XML record.  The entire XML
// table is a bunch of these linked together.  It's important to note somewhere,
// so we do it here:  Although DO-178B is a complete rewrite of DO-178A, 
// we're hoping to jam the various DO-178A documents into the same 
// document types (PSAC, SAS, etc.) and control types (CC1, CC2) as DO-178B.
struct XmlStruct
{
  // Links.
  XmlStruct *Next;
  XmlStruct *Children;
  XmlStruct *Parent;
  // Data.  
  char *Name;           // Link.
  char *Type;           // Link.
  char *Hint;           // Link.
  char *Text;           // Link.
  // The following fields relate the data contained in this item to the
  // various DO-178B documents.  Use with the constants defined above, like so:
  //    Documents[SDP] = DOCUMENT_YES;
  int Documents[NUM_DOCUMENTS];
  char Sections[NUM_DOCUMENTS][64];
  int SectionLevel[NUM_DOCUMENTS];
  // Here's where the control types for the records are indicated, 
  // vs. criticality level.  Use the constants defined above, like so:
  //    Applicability[DO178B_LEVEL_C] = SATISFIED_YES | CC_2;
  int Applicability[NUM_CRITICALITY_LEVELS];
  // Here's the wxTreeItemId given to this record in the displayed wxTreeCtrl.
  wxTreeItemId Id;
  wxColour *Color;      // Color of the item's text.
};

int LoadXmlFile (const char *Filename, struct XmlStruct *Xml);
int SaveXmlFile (const char *Filename, struct XmlStruct *Xml);
void ClearXml (struct XmlStruct *Xml);
struct XmlStruct *FindXml (struct XmlStruct *Root, wxTreeItemId Index);
int XmlToSgml (struct XmlStruct *Root, const char *OutputDirectory);

// WDR: class declarations

//----------------------------------------------------------------------------
// SearchBoxClass
//----------------------------------------------------------------------------

class SearchBoxClass:public wxDialog
{
public:
  // constructors and destructors
  SearchBoxClass (wxWindow * parent, wxWindowID id, const wxString & title,
          const wxPoint & pos = wxDefaultPosition,
          const wxSize & size = wxDefaultSize,
          long style = wxDEFAULT_DIALOG_STYLE);

  // WDR: method declarations for SearchBoxClass

  wxCheckBox *GetSearchsdd ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSDD);
  }
  wxButton *GetSearchcancel ()
  {
    return (wxButton *) FindWindow (ID_SEARCHCANCEL);
  }
  wxButton *GetSearchnow ()
  {
    return (wxButton *) FindWindow (ID_SEARCHNOW);
  }
  wxCheckBox *GetSearchall ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHALL);
  }
  wxCheckBox *GetSearchsas ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSAS);
  }
  wxCheckBox *GetSearchsci ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSCI);
  }
  wxCheckBox *GetSearchseci ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSECI);
  }
  wxCheckBox *GetSearchsvcp ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSVCP);
  }
  wxCheckBox *GetSearchsrd ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSRD);
  }
  wxCheckBox *GetSearchscs ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSCS);
  }
  wxCheckBox *GetSearchsds ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSDS);
  }
  wxCheckBox *GetSearchsrs ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSRS);
  }
  wxCheckBox *GetSearchsqap ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSQAP);
  }
  wxCheckBox *GetSearchscmp ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSCMP);
  }
  wxCheckBox *GetSearchsvp ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSVP);
  }
  wxCheckBox *GetSearchsdp ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHSDP);
  }
  wxCheckBox *GetSearchpsac ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHPSAC);
  }
  wxTextCtrl *GetSearchbox ()
  {
    return (wxTextCtrl *) FindWindow (ID_SEARCHBOX);
  }
  wxCheckBox *GetSearchhint ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHHINT);
  }
  wxCheckBox *GetSearchdescriptive ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHDESCRIPTIVE);
  }
  wxCheckBox *GetSearchtitles ()
  {
    return (wxCheckBox *) FindWindow (ID_SEARCHTITLES);
  }

private:
  // WDR: member variable declarations for SearchBoxClass

private:
  // WDR: handler declarations for SearchBoxClass
  void CheckSearchSas (wxCommandEvent & event);
  void CheckSearchSci (wxCommandEvent & event);
  void CheckSearchSeci (wxCommandEvent & event);
  void CheckSearchSvcp (wxCommandEvent & event);
  void CheckSearchSdd (wxCommandEvent & event);
  void CheckSearchSrd (wxCommandEvent & event);
  void CheckSearchScs (wxCommandEvent & event);
  void CheckSearchSds (wxCommandEvent & event);
  void CheckSearchSrs (wxCommandEvent & event);
  void CheckSearchSqap (wxCommandEvent & event);
  void CheckSearchScmp (wxCommandEvent & event);
  void CheckSearchSvp (wxCommandEvent & event);
  void CheckSearchSdp (wxCommandEvent & event);
  void CheckSearchPsac (wxCommandEvent & event);
  void CheckAllDocs (wxCommandEvent & event);
  void ClickSearchCancel (wxCommandEvent & event);
  void ClickSearchButton (wxCommandEvent & event);
  void CheckSearchHints (wxCommandEvent & event);
  void CheckSearchText (wxCommandEvent & event);
  void CheckSearchTitles (wxCommandEvent & event);

private:
DECLARE_EVENT_TABLE ()};

//----------------------------------------------------------------------------
// ProjectOptionEditorClass
//----------------------------------------------------------------------------

class ProjectOptionEditorClass:public wxDialog
{
public:
  // constructors and destructors
  ProjectOptionEditorClass (wxWindow * parent, wxWindowID id,
                const wxString & title, const wxPoint & pos =
                wxDefaultPosition, const wxSize & size =
                wxDefaultSize, long style =
                wxDEFAULT_DIALOG_STYLE);

  // WDR: method declarations for ProjectOptionEditorClass

private:
  // WDR: member variable declarations for ProjectOptionEditorClass

private:
  // WDR: handler declarations for ProjectOptionEditorClass
  //void DoneXml (wxCommandEvent & event);
  void CancelProjectOptions (wxCommandEvent & event);
  void OkProjectOptions (wxCommandEvent & event);
  void ProjectOptions (wxCommandEvent & event);

private:
  DECLARE_EVENT_TABLE ()};

//----------------------------------------------------------------------------
// XmlEditorDialogClass
//----------------------------------------------------------------------------

class XmlEditorDialogClass:public wxDialog
{
public:
  // constructors and destructors
  XmlEditorDialogClass (wxWindow * parent, wxWindowID id,
            const wxString & title, const wxPoint & pos =
            wxDefaultPosition, const wxSize & size =
            wxDefaultSize, long style = wxDEFAULT_DIALOG_STYLE);

  // WDR: method declarations for XmlEditorDialogClass
    wxBitmapButton* GetBuildbutton()  { return (wxBitmapButton*) FindWindow( ID_BUILDBUTTON ); }
    wxBitmapButton* GetInsertMdash()  { return (wxBitmapButton*) FindWindow( ID_INSERT_MDASH ); }
    wxBitmapButton* GetInsertFootnote()  { return (wxBitmapButton*) FindWindow( ID_INSERT_FOOTNOTE ); }
    wxBitmapButton* GetInsertHyperlink()  { return (wxBitmapButton*) FindWindow( ID_INSERT_HYPERLINK ); }
    wxBitmapButton* GetSubitemCustom()  { return (wxBitmapButton*) FindWindow( ID_SUBITEM_CUSTOM ); }
    wxBitmapButton* GetAddItemCustom()  { return (wxBitmapButton*) FindWindow( ID_ADD_ITEM_CUSTOM ); }
  //wxChoice* GetDocchoice()  { return (wxChoice*) FindWindow( ID_DOCCHOICE ); }
  wxBitmapButton *GetInsertGraphic ()
  {
    return (wxBitmapButton *) FindWindow (ID_INSERT_GRAPHIC);
  }
  wxBitmapButton *GetInsertList ()
  {
    return (wxBitmapButton *) FindWindow (ID_INSERT_LIST);
  }
  wxBitmapButton *GetInsertTable ()
  {
    return (wxBitmapButton *) FindWindow (ID_INSERT_TABLE);
  }
  wxBitmapButton *GetDocnext ()
  {
    return (wxBitmapButton *) FindWindow (ID_DOCNEXT);
  }
  wxChoice *GetDocchoice ()
  {
    return (wxChoice *) FindWindow (ID_DOCCHOICE);
  }
  wxBitmapButton *GetDocback ()
  {
    return (wxBitmapButton *) FindWindow (ID_DOCBACK);
  }
  wxStaticLine *GetLine ()
  {
    return (wxStaticLine *) FindWindow (ID_LINE);
  }
  wxStaticText *GetAppE ()
  {
    return (wxStaticText *) FindWindow (ID_APP_E);
  }
  wxStaticText *GetAppD ()
  {
    return (wxStaticText *) FindWindow (ID_APP_D);
  }
  wxStaticText *GetAppC ()
  {
    return (wxStaticText *) FindWindow (ID_APP_C);
  }
  wxStaticText *GetAppB ()
  {
    return (wxStaticText *) FindWindow (ID_APP_B);
  }
  wxStaticText *GetAppA ()
  {
    return (wxStaticText *) FindWindow (ID_APP_A);
  }
  wxStaticText *GetAppLevel ()
  {
    return (wxStaticText *) FindWindow (ID_APP_LEVEL);
  }
  wxBitmapButton *GetDockbook ()
  {
    return (wxBitmapButton *) FindWindow (ID_DOCKBOOK);
  }
  //wxBitmapButton* GetBuildbutton()  { return (wxBitmapButton*) FindWindow( ID_BUILDBUTTON ); }
  wxBitmapButton *GetSubitem ()
  {
    return (wxBitmapButton *) FindWindow (ID_SUBITEM);
  }
  wxCheckBox *GetSvpsection ()
  {
    return (wxCheckBox *) FindWindow (ID_SVPSECTION);
  }
  wxBitmapButton *GetDeleteItem ()
  {
    return (wxBitmapButton *) FindWindow (ID_DELETE_ITEM);
  }
  wxBitmapButton *GetAddItem ()
  {
    return (wxBitmapButton *) FindWindow (ID_ADD_ITEM);
  }
  wxCheckBox *GetSassection ()
  {
    return (wxCheckBox *) FindWindow (ID_SASSECTION);
  }
  //wxCheckBox* GetSqarsection()  { return (wxCheckBox*) FindWindow( ID_SQARSECTION ); }
  //wxCheckBox* GetScmrsection()  { return (wxCheckBox*) FindWindow( ID_SCMRSECTION ); }
  wxCheckBox *GetScisection ()
  {
    return (wxCheckBox *) FindWindow (ID_SCISECTION);
  }
  wxCheckBox *GetSecisection ()
  {
    return (wxCheckBox *) FindWindow (ID_SECISECTION);
  }
  //wxCheckBox* GetSvrsection()  { return (wxCheckBox*) FindWindow( ID_SVRSECTION ); }
  wxCheckBox *GetSvcpsection ()
  {
    return (wxCheckBox *) FindWindow (ID_SVCPSECTION);
  }
  wxCheckBox *GetDdsection ()
  {
    return (wxCheckBox *) FindWindow (ID_DDSECTION);
  }
  wxCheckBox *GetSrdsection ()
  {
    return (wxCheckBox *) FindWindow (ID_SRDSECTION);
  }
  wxCheckBox *GetScssection ()
  {
    return (wxCheckBox *) FindWindow (ID_SCSSECTION);
  }
  wxCheckBox *GetSdssection ()
  {
    return (wxCheckBox *) FindWindow (ID_SDSSECTION);
  }
  wxCheckBox *GetSrssection ()
  {
    return (wxCheckBox *) FindWindow (ID_SRSSECTION);
  }
  wxCheckBox *GetSqapsection ()
  {
    return (wxCheckBox *) FindWindow (ID_SQAPSECTION);
  }
  wxCheckBox *GetScmpsection ()
  {
    return (wxCheckBox *) FindWindow (ID_SCMPSECTION);
  }
  wxCheckBox *GetSdpsection ()
  {
    return (wxCheckBox *) FindWindow (ID_SDPSECTION);
  }
  wxCheckBox *GetPsacsection ()
  {
    return (wxCheckBox *) FindWindow (ID_PSACSECTION);
  }
  wxCheckBox *GetSascheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SASCHECK);
  }
  //wxCheckBox* GetSqarcheck()  { return (wxCheckBox*) FindWindow( ID_SQARCHECK ); }
  //wxCheckBox* GetScmrcheck()  { return (wxCheckBox*) FindWindow( ID_SCMRCHECK ); }
  wxCheckBox *GetScicheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SCICHECK);
  }
  wxCheckBox *GetSecicheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SECICHECK);
  }
  //wxCheckBox* GetSvrcheck()  { return (wxCheckBox*) FindWindow( ID_SVRCHECK ); }
  wxCheckBox *GetSvcpcheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SVCPCHECK);
  }
  wxCheckBox *GetDdcheck ()
  {
    return (wxCheckBox *) FindWindow (ID_DDCHECK);
  }
  wxCheckBox *GetSrdcheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SRDCHECK);
  }
  wxCheckBox *GetScscheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SCSCHECK);
  }
  wxCheckBox *GetSdscheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SDSCHECK);
  }
  wxCheckBox *GetSrscheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SRSCHECK);
  }
  wxCheckBox *GetSqapcheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SQAPCHECK);
  }
  wxCheckBox *GetScmpcheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SCMPCHECK);
  }
  wxCheckBox *GetSvpcheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SVPCHECK);
  }
  wxCheckBox *GetSdpcheck ()
  {
    return (wxCheckBox *) FindWindow (ID_SDPCHECK);
  }
  wxCheckBox *GetPsaccheck ()
  {
    return (wxCheckBox *) FindWindow (ID_PSACCHECK);
  }
  wxTextCtrl *GetHinttext ()
  {
    return (wxTextCtrl *) FindWindow (ID_HINTTEXT);
  }
  wxTextCtrl *GetItemtext ()
  {
    return (wxTextCtrl *) FindWindow (ID_ITEMTEXT);
  }
  wxTreeCtrl *GetXmleditor ()
  {
    return (wxTreeCtrl *) FindWindow (ID_XMLEDITOR);
  }
  wxBitmapButton *GetCce ()
  {
    return (wxBitmapButton *) FindWindow (ID_CCE);
  }
  wxBitmapButton *GetCcd ()
  {
    return (wxBitmapButton *) FindWindow (ID_CCD);
  }
  wxBitmapButton *GetCcc ()
  {
    return (wxBitmapButton *) FindWindow (ID_CCC);
  }
  wxBitmapButton *GetCcb ()
  {
    return (wxBitmapButton *) FindWindow (ID_CCB);
  }
  wxBitmapButton *GetCca ()
  {
    return (wxBitmapButton *) FindWindow (ID_CCA);
  }
  wxBitmapButton *GetSate ()
  {
    return (wxBitmapButton *) FindWindow (ID_SATE);
  }
  wxBitmapButton *GetSatc ()
  {
    return (wxBitmapButton *) FindWindow (ID_SATC);
  }
  wxBitmapButton *GetSatd ()
  {
    return (wxBitmapButton *) FindWindow (ID_SATD);
  }
  wxBitmapButton *GetSatb ()
  {
    return (wxBitmapButton *) FindWindow (ID_SATB);
  }
  wxBitmapButton *GetSata ()
  {
    return (wxBitmapButton *) FindWindow (ID_SATA);
  }
  wxBitmapButton *GetFindagainbutton ()
  {
    return (wxBitmapButton *) FindWindow (ID_FINDAGAINBUTTON);
  }
  wxRadioButton *GetExpertbutton ()
  {
    return (wxRadioButton *) FindWindow (ID_EXPERTBUTTON);
  }
  wxRadioButton *GetNovicebutton ()
  {
    return (wxRadioButton *) FindWindow (ID_NOVICEBUTTON);
  }
  wxBitmapButton *GetMarkupbutton ()
  {
    return (wxBitmapButton *) FindWindow (ID_MARKUPBUTTON);
  }
  wxChoice *GetMarkupchoice ()
  {
    return (wxChoice *) FindWindow (ID_MARKUPCHOICE);
  }

public:
  void FindAgainButton (wxCommandEvent & event);
  void ClickCleanup (struct XmlStruct *Record);
  struct XmlStruct *RootXml, *Template;
  int ClickTreeBypass;
  wxTreeItemId LastSelected;
  wxTreeCtrl *ItemTree;
  void OnClose (wxCloseEvent & event);
  void OnSize (wxSizeEvent & event);
  void XmlChange (wxTreeItemId Selected, struct XmlStruct *Xml);
  void FrameworkSet (struct XmlStruct *Xml);
  int DatasetNewlyChanged;
  void DocumentChange (struct XmlStruct *, int DocNum, wxCheckBox * Box);
  void SatChange (struct XmlStruct *, int Level, wxBitmapButton * Button);
  void CcChange (struct XmlStruct *, int Level, wxBitmapButton * Button);
  void ButtonBitmap (wxBitmapButton * Button, wxBitmap Bitmap);
  void SetSectionDocs (struct XmlStruct *Record);
  struct XmlStruct **FindPredecessor (struct XmlStruct *Root,
                      struct XmlStruct *Xml);
  void ClickTree (wxTreeEvent & event);
  // Stuff related to search parameters.
  wxString SearchString;
  bool SearchDocs[NUM_DOCUMENTS];
  bool SearchAll, SearchTitles, SearchHints, SearchText;
  wxTreeItemId LastSearchPos;

private:
  // WDR: member variable declarations for XmlEditorDialogClass

  // The color used to mark data items changed.
  wxColour MY_CHANGE_COLOR;
  wxColour *CHANGE_COLOR;


// These are pointers to various interesting widgets within an XML-editor
// dialog.  they are valid only while the dialog exists, and must be 
// regenerated each time the dialog starts.
  wxTextCtrl *HintBox, *ItemText;
  wxCheckBox *PsacCheck, *SdpCheck, *SvpCheck, *ScmpCheck, *SqapCheck,
    *SrsCheck, *SdsCheck, *ScsCheck, *SrdCheck, *DdCheck,
    *SvcpCheck, *SeciCheck, *SciCheck, *SasCheck;
  wxCheckBox *PsacSection, *SdpSection, *SvpSection, *ScmpSection,
    *SqapSection, *SrsSection, *SdsSection, *ScsSection, *SrdSection,
    *DdSection, *SvcpSection, *SeciSection, *SciSection, *SasSection;
  wxBitmapButton *BitmapSatA, *BitmapSatB, *BitmapSatC, *BitmapSatD,
    *BitmapSatE, *CcA, *CcB, *CcC, *CcD, *CcE,
    *InsertGraphic, *InsertList, *InsertTable,
    *InsertHyperlink, *InsertFootnote, *InsertMdash;
  wxButton *Add, *AddSub, *Del;
  wxButton *AddCustom, *AddSubCustom;
  wxChoice *MarkupChoice, *DocChoice;
  wxBitmapButton *MarkupButton;
  wxRadioButton *NoviceButton, *ExpertButton;
  wxStaticText *AppLevel, *AppA, *AppB, *AppC, *AppD, *AppE;
  wxStaticLine *SeparatorLine;

private:
  void SetAndCheckSelection (long from, long to);
  // WDR: handler declarations for XmlEditorDialogClass
    void ClickBuildButton( wxCommandEvent &event );
    void InsertMdashClick( wxCommandEvent &event );
    void InsertFootnoteClick( wxCommandEvent &event );
    void InsertHyperlinkClick( wxCommandEvent &event );
    void AddSubitemCustomClick( wxCommandEvent &event );
    void AddItemCustomClick( wxCommandEvent &event );
  void ClickInsertGraphic (wxCommandEvent & event);
  void ClickInsertList (wxCommandEvent & event);
  void ClickInsertTable (wxCommandEvent & event);
  void ClickDocNext (wxCommandEvent & event);
  void ClickDocBack (wxCommandEvent & event);
  void TraceClick (wxCommandEvent & event);
  void DockBookButtonClick (wxCommandEvent & event);
  void BuildButtonClick (wxCommandEvent & event);
  void ClickListingButton (wxCommandEvent & event);
  void ClickCodeButton (wxCommandEvent & event);
  void ClickCommandButton (wxCommandEvent & event);
  void ClickEmphasisButton (wxCommandEvent & event);
  void ExpertCheck (wxCommandEvent & event);
  void NoviceCheck (wxCommandEvent & event);
  void MarkupButtonClick (wxCommandEvent & event);
  void FindButton (wxCommandEvent & event);
  struct XmlStruct *SeekTemplate (struct XmlStruct *Record);
  struct XmlStruct *CreateTreeCopy (struct XmlStruct *Xml);
  void FillSubTree (struct XmlStruct *Xml);
  struct XmlStruct *AllowableAddSiblings (struct XmlStruct *Record);
  struct XmlStruct *AllowableAddSubs (struct XmlStruct *Record);
  void SqarClick (wxCommandEvent & event);
  void ScmrClick (wxCommandEvent & event);
  void SvrClick (wxCommandEvent & event);
  void SaveButton (wxCommandEvent & event);
  void DeleteItem (wxCommandEvent & event);
  void AddSubItem (wxCommandEvent & event);
  void AddItem (wxCommandEvent & event);
  void ClickCcE (wxCommandEvent & event);
  void ClickCcD (wxCommandEvent & event);
  void ClickCcC (wxCommandEvent & event);
  void ClickCcB (wxCommandEvent & event);
  void ClickCcA (wxCommandEvent & event);
  void ClickSatE (wxCommandEvent & event);
  void ClickSatD (wxCommandEvent & event);
  void ClickSatC (wxCommandEvent & event);
  void ClickSatB (wxCommandEvent & event);
  void ClickSatA (wxCommandEvent & event);
  void SasClick (wxCommandEvent & event);
  void SciClick (wxCommandEvent & event);
  void SeciClick (wxCommandEvent & event);
  void SvcpClick (wxCommandEvent & event);
  void DdClick (wxCommandEvent & event);
  void SrdClick (wxCommandEvent & event);
  void ScsClick (wxCommandEvent & event);
  void SdsClick (wxCommandEvent & event);
  void SrsClick (wxCommandEvent & event);
  void SqapClick (wxCommandEvent & event);
  void ScmpClick (wxCommandEvent & event);
  void SvpClick (wxCommandEvent & event);
  void SdpClick (wxCommandEvent & event);
  void PsacClick (wxCommandEvent & event);
  //void DoneXml (wxCommandEvent & event);

private:
DECLARE_EVENT_TABLE ()};

//----------------------------------------------------------------------------
// Do178BuilderFrame
//----------------------------------------------------------------------------

class Do178BuilderFrame:public wxFrame
{
public:
  // constructors and destructors
  Do178BuilderFrame (wxWindow * parent, wxWindowID id, const wxString & title,
             const wxPoint & pos = wxDefaultPosition,
             const wxSize & size = wxDefaultSize,
             long style = wxDEFAULT_FRAME_STYLE);

public:
  int FileIsOpen;
  void OpenTemplate (void);

private:
  // WDR: method declarations for Do178BuilderFrame
  void CreateMyMenuBar ();
  void CreateMyToolBar ();
  void LoadXmlToEditor (struct XmlStruct *, XmlEditorDialogClass &,
            wxTreeItemId);

private:
  // WDR: member variable declarations for Do178BuilderFrame

private:
  // WDR: handler declarations for Do178BuilderFrame
  void PruneD (wxCommandEvent & event);
  void PruneC (wxCommandEvent & event);
  void PruneB (wxCommandEvent & event);
  void OnNew (wxCommandEvent & event); 
  void BuildDocuments (wxCommandEvent & event);
  void SaveFileAs (wxCommandEvent & event);
  void SaveFile (wxCommandEvent & event);
  void CloseFile (wxCommandEvent & event);
  void HintEditCheck (wxCommandEvent & event);
  void EditDataset (wxCommandEvent & event);
  void ProjectOptions (wxCommandEvent & event);
  void Open (wxCommandEvent & event);
  //void OpenMenu( wxCommandEvent &event );
  void OnAbout (wxCommandEvent & event);
  void OnQuit (wxCommandEvent & event);
  void OnCloseWindow (wxCloseEvent & event);

private:
  int UnsavedCancel (void);

private:
  DECLARE_EVENT_TABLE ()};

//----------------------------------------------------------------------------
// Do178BuilderApp
//----------------------------------------------------------------------------

class Do178BuilderApp:public wxApp
{
public:
  Do178BuilderApp ();

  virtual bool OnInit ();
  virtual int OnExit ();
};

#endif
