////////////////////////////////////////////////////////////////////////////
// Filename:    LoadXmlFile.cpp
// Purpose:     Loads an Do178Builder XML file.
// Mods:        09/23/01 RSB    Began.
//              10/13/01 RSB    Added traceability.
//              10/21/01 RSB    Added control of output directory.
//                              Now ignores TBD entries in document 
//                              production.  Parses and includes the header
//                              info now.
//              10/22/01 RSB    Debugged printout of multi-paragraph text.
//                              Eliminated the automatic removal of "TBD"
//                              sections, since if those are auto-eliminated
//                              it tends to give the false impression that
//                              the documents are more complete than they
//                              really are.
//              10/29/01 RSB    Now allow the DocBook tags <funcname>,
//                              <type>, <varname>, and <programlisting>
//                              to be written to the sgml files.
//                              Fixed bug in which &#nn; misinterpreted
//                              on input.
//              10/30/01 RSB    Added DocBook <emphasis> and 
//                              </emphasis> and a lot of others, too.
//              10/31/01 RSB    Now recognizes all DocBook 3.1 inline tags.
//                              Added backup files on save.
//                              Fixed serious bug in which TYPE is saved with
//                              spaces rather than underlines, and hence is
//                              discarded on later loading.
//              02/17/02 RSB    Corrected interpretation of '&', '<', '>' in
//                              the input XML file.
//              02/25/02 RSB    Fixed the article-header graphic.
//                              Removed ForeignPhrase markup.
//              11/10/03 RSB    Now allows for separate numbers/revisions
//                              for the various documents, through the 
//                              XxxxIdentifier field.  Also, graphic files
//                              for figures are copied into the output-doc
//                              directory, and referenced from there from
//                              within the output SGML files.  A log file is
//                              created when SGML docs are created, allowing
//                              SGML sections to be traced to XML sections.
//              11/11/03 RSB    A copy of the XML file called Sections.xml
//                              is now saved during SGML creation.
//              03/26/04 RSB    Added the <DocBook> tag.
//              03/28/04 RSB    Added code to fetch graphics (or other files)
//                              in <DocBook>-encapsulated filerefs into
//                              the output directory.
//              04/17/04 RSB    Added the <trace> tag for putting additional
//                              traceability refs in the SRD, SDD, or SVCP.
//              05/15/06 RSB    Error messages pepped up a little (they're
//                              still might rotten!) and increased the size
//                              of a field from 32K to 64K.
//              06/23/06 RSB    Began adapting for DO-254.
//              06/26/06 RSB    Allowed for traceability from high-level 
//                              requirements to test cases.  Multiple 
//                              traceability refs on a single item now
//                              displayed better.
//              06/27/06 RSB    Had to tweak the traceability to get it to
//                              work in all the new ways I want (SRD->SVCP).
//              06/29/06 RSB    Upon loading a database, now automatically
//                              adjusts the Applicabiilty entries so that
//                              a parent item's applicability is at least
//                              that of all the children.  Added function to
//                              prune to a given assurance level.
//              03/12/07 RSB    NoForwardTrace added to eliminate forward
//                              traceability references.  Also, NoTrace.
//              01/22/08 RSB    Added --combine-srd-sdd.
//              01/25/08 RSB    Fixed the "release" field on the title page
//                              for combined SRD/SDD documents.
//              01/26/08 RSB    Added DocBook 4.5 XML output.
//              01/27/08 RSB    References to '<' and '&' (rather than
//                              "&lt;" and "&amp;") in <programlisting> that
//                              SGML sometimes accepts but XML never accepts
//                              are now fixed.  Also, now allow for DocBook
//                              v4.1.2, v4.2, v4.3, and v4.4.
//              01/29/08 RSB    Upped the limit again (see 05/15/06 above)
//                              again on the size of a text field.  Now
//                              reallocates automatically.
//              01/29/08 RSB    (Later.)  Found that the new automatic
//                              attribute quoting I've added within
//                              <docbook></docbook> tags will hang if
//                              there's an unterminated tag.  I've fixed this.
//                              Also, in case the problem is a newline in place
//                              of a closing '>', added some helpful error
//                              messages.  Cleaned up the messages in
//                              general that are displayed on the console.
//              03/08/08 RSB    Accounted for possible missing closing tag for
//                              <imagedata> when using --x4 switches.
//                              Added NoProductname (--no-productname).
//              03/13/08 RSB    Simplified zlib usage.
//              03/26/08 RSB    Fixed up for wxWidgets 2.8.
//              04/02/08 RSB    Made some fixes that ease up using docbook
//                              markup that has been created by exporting
//                              from OpenOffice.  Fixed the problem that if
//                              both --combine-srd-sdd and --DO-254 are used
//                              the combined HRD/HDRD is named as if it was
//                              simply the HRD by itself.
//              04/03/08 RSB    My homegrown parsing is really suckified.
//                              Added handling of SGML/XML comments within
//                              a <docbook>...</docbook> envelope.
//              04/05/08 RSB    Now passes markup like <?...?> through 
//                              unchanged to the output file.  Removed some
//                              additional possibilities for pernicious 
//                              mixed content.
//		01/28/09 RSB	Fixed lots of compiler warnings/errors that 
//				seem to be associated with using Unicode
//				builds of wxWidgets.
//              02/03/09 RSB    Added stuff for --combine-plans.

/*
  Copyright 2001-2004,2006-2009 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

////////////////////////////////////////////////////////////////////////////

// For compilers that support precompilation
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

// Include private headers
#include "Do178Builder.h"

// The zlib library is used to generate the CRCs that are inserted into the
// Sections.log file, and for nothing else.  Some people have had trouble 
// getting it to compile in the Windows environment.  Since the licensing
// of zlib is such as to allow it, I've solved that problem by simply 
// sucking zlib/crc32.c unchanged into the Do178Builder project, and then
// providing enough of a replacement for zlib.h to account for the datatypes
// needed.  So the line below is my custom zlib.h.
#include "zlib.h"

// Names of Unicode entities suitable for DocBook.
#include "entities.h"

extern int NoForwardTrace, NoTrace, CombineSrdSdd, x4;
extern char *CombinedPlanName;
extern char *x4v1, *x4v2;

//-------------------------------------------------------------------------
// 06/29/2006 RSB.
// Normalize a hierarchical dataset so that the applicability of any
// given item is at least as high as the applicability of any of its 
// children.  This is necessary when pruning the tree for a given
// assurance level.  

void
NormalizeApplicability (struct XmlStruct *Root)
{
  extern int Do254;
  static wxColour AppChangedColour;
  struct XmlStruct *Next;
  static int RecursionLevel = -1;
  int i;

  if (Root == NULL)
    return;
  RecursionLevel++;
  // Loop on the children, and adjust Root->Applicability[]
  // upward if needed.
  if (Root->Children == NULL && Do254)
    {
      // For DO-254, get rid of the "satisfaction" fields.
      for (i = 0; i < NUM_CRITICALITY_LEVELS; i++)
	Root->Applicability[i] &= CC_MASK;
    }
  else
    {
      for (Next = Root->Children; Next != NULL; Next = Next->Next)
	{
	  // Obtain the normalized Applicability[] of the child.
	  NormalizeApplicability (Next);
	  // Upgrade the root Applicability[] if the child's is greater.
	  // We don't both with changing the true root, since it's
	  // fictitious.
	  if (RecursionLevel > 0)	// Not true root?
	    {
	      for (i = 0; i < NUM_CRITICALITY_LEVELS; i++)
		{
		  int CcRoot, CcChild, SatisfiedRoot, SatisfiedChild;
		  int NewApplicability;
		  CcRoot = Root->Applicability[i] & CC_MASK;
		  CcChild = Next->Applicability[i] & CC_MASK;
		  if (CcChild > CcRoot)
		    CcRoot = CcChild;
		  if (Do254)
		    {
		      SatisfiedRoot = SatisfiedChild = SATISFIED_DISCRETION;
		    }
		  else
		    {
		      SatisfiedRoot = Root->Applicability[i] & SATISFIED_MASK;
		      SatisfiedChild =
			Next->Applicability[i] & SATISFIED_MASK;
		      if (SatisfiedChild > SatisfiedRoot)
			SatisfiedRoot = SatisfiedChild;
		    }
		  NewApplicability = CcRoot | SatisfiedRoot;
		  if (NewApplicability != Root->Applicability[i])
		    {
		      Root->Applicability[i] = NewApplicability;
		      AppChangedColour = wxColour (0, 255, 0);
		      Root->Color = &AppChangedColour;
		    }
		}
	    }
	}
    }
  RecursionLevel--;
}

//------------------------------------------------------------------
// Given the name of a graphic in the current directory, copies its
// file into the output directory.  Unfortunately, because of the 
// nature of DocBook, we can't know exactly what the filename of the
// graphic is.  For example, if the graphic name is "MyGraphic", the
// graphic's filename might be MyGraphic.jpg, MyGraphic.eps, etc.
// These are actually the two likeliest cases, so what we do is to 
// copy BOTH of those (and maybe others in the future).

static const char *GraphicExtensions[] = { ".jpg", ".eps", "" };

#define NUM_EXTENSIONS (sizeof (GraphicExtensions) / sizeof (char *))
static char FileBuffer[16284];

static void
FetchGraphic (char *GraphicName, const char *OutputDirectory)
{
  char *InName, *OutName;
  FILE *fin, *fout;
  unsigned i;
  int j, k;
  InName = (char *) malloc (5 + strlen (GraphicName));
  if (InName == NULL)
    return;
  OutName =
    (char *) malloc (5 + strlen (GraphicName) + strlen (OutputDirectory));
  if (OutName == NULL)
    {
      free (InName);
      return;
    }
  for (i = 0; i < NUM_EXTENSIONS; i++)
    {
      sprintf (InName, "%s%s", GraphicName, GraphicExtensions[i]);
      sprintf (OutName, "%s%s", OutputDirectory, InName);
      fin = fopen (InName, "rb");
      if (fin == NULL)
	continue;
      fout = fopen (OutName, "wb");
      if (fout == NULL)
	{
	  fclose (fin);
	  break;
	}
      for (k = 0; 0 < (j = fread (FileBuffer, 1, sizeof (FileBuffer), fin));
	   k += j)
	fwrite (FileBuffer, 1, j, fout);
      printf ("%s -> %s, %d bytes.\n", InName, OutName, k);
      fclose (fout);
      fclose (fin);
    }
  free (OutName);
  free (InName);
}

//------------------------------------------------------------------
// Clears an XML list in memory.  All linked records and their
// linked data are deallocated.  The root of the linked list, however,
// is merely cleared, and no attempt is made to deallocate it.

void
ClearXml (struct XmlStruct *Xml)
{
  struct XmlStruct *Next, *NextNext;
  for (Next = Xml; Next != NULL;)
    {
      if (Next->Children != NULL)
	{
	  ClearXml (Next->Children);
	  // Remember, ClearXml won't deallocate the root of 
	  // Next->Children, so we have to do that explicitly.
	  free (Next->Children);
	  Next->Children = NULL;
	}
      if (NULL != Next->Name)
	{
	  free (Next->Name);
	  Next->Name = NULL;
	}
      if (NULL != Next->Hint)
	{
	  free (Next->Hint);
	  Next->Hint = NULL;
	}
      if (NULL != Next->Text)
	{
	  free (Next->Text);
	  Next->Text = NULL;
	}
      // Deallocate this entire record -- unless it's the root. 
      NextNext = Next->Next;
      if (Next != Xml)
	free (Next);
      else
	Next->Next = NULL;
      Next = NextNext;
    }
}

//------------------------------------------------------------------
// Converts a tag name to something nicely displayable, by converting
// underlines to spaces, and some other conversions as well.

static void
TagToName (char *Text)
{
  char c, *s;
  int i;
  while (*Text)
    {
      if (*Text == '_')
	*Text++ = ' ';
      else if (*Text == '&' && 2 == sscanf (Text, "&#%d%c", &i, &c)
	       && c == ';')
	{
	  if (i == '\n' || (i >= ' ' && i < 127))
	    *Text++ = i;
	  else
	    *Text++ = '.';
	  for (s = Text; *s != ';'; s++);
	  strcpy (Text, s + 1);
	}
      else
	Text++;
    }
}

// ... and vice-versa.

void
NameToTag (const char *Name, char *Tag)
{
  while (*Name)
    {
      if (*Name == ' ')
	{
	  *Tag++ = '_';
	  Name++;
	}
      else if (*Name == '-' || isalnum (*Name))
	*Tag++ = *Name++;
      else
	{
	  sprintf (Tag, "&#%d;", *Name++);
	  Tag += strlen (Tag);
	}
    }
  *Tag = 0;
}

//------------------------------------------------------------------
// Reads a file until a markup tag is found, and returns the tag.
// The return value is 0 on success.  Call the function with 
// TagSize equal to sizeof(Tag), where Tag has been allocated as
// char Tag[something].  Note that this code won't work if the 
// quoted strings in the tag contain '>' or '<'.

static int
GetTag (FILE * fp, char *Tag, int TagSize)
{
  int ch;
  for (;;)
    {
      ch = getc (fp);
      if (ch == EOF)
	return (1);
      if (ch == '<')		// Found start of tag.
	{
	  for (;;)
	    {
	      ch = getc (fp);
	      if (ch == EOF)
		return (1);
	      if (ch == '>')	// Found end of tag.
		{
		  if (TagSize <= 0)
		    return (1);
		  *Tag = '\0';
		  return (0);
		}
	      else
		{
		  if (TagSize <= 0)
		    return (1);
		  *Tag++ = ch;
		  TagSize--;
		}
	    }
	}
    }
}

//------------------------------------------------------------------
// Read textual data between tags.  Return:
//      0 success
//      1 buffer too small
//      2 file-read error.

static int
GetBetweenTag (FILE * fp, char *Text, int TextSize)
{
  int ch;
  char c, *s;
  int i;
  s = Text;
  for (;;)
    {
      ch = getc (fp);
      if (ch == EOF)
	return (2);
      if (ch == '<')		// Found start of tag.
	{
	  ungetc (ch, fp);
	  if (TextSize <= 0)
	    return (1);
	  *Text = '\0';
	  break;
	}
      else
	{
	  if (TextSize <= 0)
	    return (1);
	  *Text++ = ch;
	  TextSize--;
	}
    }
  // Convert things like &#60; &#62; &#38; back to < > &.
  for (; *s; s++)
    if (2 == sscanf (s, "&#%d%c", &i, &c) && c == ';' &&
	((i >= ' ' && i < 127) || (i >= 160 && i <= 255)))
      {
	*s = i;
	strcpy (s + 1, s + 5 + (i > 99));
      }
  return (0);
}

//------------------------------------------------------------------
// Cleans up the text read from the input file in various ways, 
// such as trimming leading/trailing space and translating
// special characters like '<' and '>'.

static void
CleanupText (char *s)
{
  char *ss, *sss = NULL, c;
  for (ss = s; isspace (*ss); ss++);
  while (*ss)
    {
      if (*ss == '&')
	{
	  if (!strncasecmp (ss, "&amp;", 5) || !strncasecmp (ss, "&#38;", 5))
	    {
	      c = '&';
	      ss += 5;
	    }
	  else if (!strncasecmp (ss, "&gt;", 4))
	    {
	      c = '>';
	      ss += 4;
	    }
	  else if (!strncasecmp (ss, "&#62;", 5))
	    {
	      c = '>';
	      ss += 5;
	    }
	  else if (!strncasecmp (ss, "&lt;", 4))
	    {
	      c = '<';
	      ss += 4;
	    }
	  else if (!strncasecmp (ss, "&#60;", 5))
	    {
	      c = '<';
	      ss += 5;
	    }
	  else
	    c = *ss++;
	}
      else
	c = *ss++;
      *s++ = c;
      if (!isspace (c))
	sss = s;
    }
  if (sss == NULL)
    *s = '\0';
  else
    *sss = '\0';
}

//------------------------------------------------------------------
// Interprets the meanings of the applicability/cc fields in the tags.

static int
InterpretLevel (char *s)
{
  int RetVal;
  switch (*s)
    {
    case 'N':
      RetVal = SATISFIED_DISCRETION;
      break;
    case 'Y':
      RetVal = SATISFIED_YES;
      break;
    case 'I':
      RetVal = SATISFIED_INDEPENDENCE;
      break;
    default:
      return (0);
    }
  switch (s[1])
    {
    case 'N':
      RetVal |= CC_DISCRETION;
      break;
    case '1':
      RetVal |= CC_1;
      break;
    case '2':
      RetVal |= CC_2;
      break;
    default:
      return (0);
    }
  return (RetVal);
}

//---------------------------------------------------------------------
// Interprets the meanings of the document mappings.

static int
InterpretDoc (char *s)
{
  if (0 == strcasecmp (s, "YES"))
    return (DOCUMENT_YES);
  else
    return (DOCUMENT_DISCRETION);
}

//------------------------------------------------------------------
// Recursive function for adding trees gotten from the input XML
// file.  On the first entry, Level must be 0.  Returns 0 on success.
// Expects a tree to end with a tag like </something>, except if 
// Level == 0, in which case the end of file terminates.  Also, 
// Xml should already have been loaded with the root entry, and the
// opening tag for the root entry should have been removed from the
// input file. 

#define MAX_TAGSIZE 4096
#define MAX_TAGFIELDS 32
static char Tag[MAX_TAGSIZE];
static char TagFields[MAX_TAGFIELDS][MAX_TAGSIZE];
static char ActualTag[128];

long LowLevelError = -1;
int LowReturnCode;
static int
LoadXmlLevel (int Level, FILE * fp, struct XmlStruct *Xml)
{
  static char *Text = NULL;	//Text[65536 /*32768 */ ];
  static long int TextFieldSize = 0;
  long Where, LowWhere;
  struct XmlStruct *NewXml, Default = { NULL }, *Root;
  char *s, *ss;
  int ReturnValue = 1, i, Hinted;
  int Position, NumInLevel, TagFieldCount;
  if (Level == 0)
    LowLevelError = -1;
  Where = ftell (fp);
  Level++;
  Root = Xml;
  Default.Color = (wxColour *) wxBLACK;	// Default text color black.
  for (NumInLevel = 0;; Xml = NewXml, NumInLevel++)
    {
      // Read the next tag.  If none found, it's an error unless
      // Level == 0, in which case it's a natural exit.
      if (GetTag (fp, Tag, sizeof (Tag)))
	{
	  // Maybe error exit, maybe not.
	  ReturnValue = 2 * (Level != 0);
	  goto Error;
	}
      // Okay, we've found a tag.  If it's an end-tag, return.
      if (Tag[0] == '/')
	goto Fine;
      // If this is the first tag we've found on this level, we need to
      // add a child to the root.  Otherwise, we need to add a sibling
      // to the last element added.
      NewXml = (struct XmlStruct *) malloc (sizeof (struct XmlStruct));
      if (NewXml == NULL)
	{
	  ReturnValue = 3;
	  goto Error;
	}
      *NewXml = Default;
      NewXml->Parent = Root;
      if (NumInLevel == 0)
	Xml->Children = NewXml;
      else
	Xml->Next = NewXml;
      // Parse the start-tag.  The format is this:
      //        <tag var1="data1" var2="data2" ... varN="dataN">
      // The "tag", "varN", and "dataN" strings do not contain spaces,
      // nor the characters '<', '>', or '"'.  In fact, all special 
      // characters other than '-' or alpha-numerics are represented by
      // numeric codes of the form "&#n;".  On input, '_' is also 
      // translated to ' '.  Here are the recognized variables:
      //        HINT, 
      //        TYPE,
      //        PSAC, SDP, ... (other docs) ..., SAS,
      //        A1, A2, A3, BA, BB, BC, BD, BE
      // Though technically any data strings meeting the conditions
      // described above are acceptable, the interpretations of the
      // variables actually limit what may be found:
      //        HINT    Arbitrary text.  Gives the user a hint as to
      //                what this particular data item is used for.
      //        TYPE    Limit to alphanumerics, '-', and '_'.
      //                This is by default the same as the tag, 
      //                but after user-editing the tag may change.
      //                The TYPE never changes.
      //        PSAC..SAS       These are used to determine whether 
      //                the particular type of data item appears
      //                in the indicated document type in general.
      //                In practice, whether it appears or not
      //                also depends on the criticality level.
      //                The allowed values are "NO" (the default)
      //                and "YES".
      //        A1..BE  These are used to the "applicability"
      //                and "control category" at the various
      //                criticality levels.  (Refer to DO-178B
      //                table A-1.)   The allowed values consist
      //                of two 1-character fields, the first 
      //                representing the "applicability":
      //                        N       "Applicant's discretion"
      //                        Y       Yes, applicable.
      //                        I       Yes, applicable with "independence".
      //                The second 1-character field determines
      //                the "control category":
      //                        N       "Applicant's discretion"
      //                        1       CC1
      //                        2       CC2
      TagFieldCount = sscanf (Tag, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s"
			      "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
			      ActualTag,
			      TagFields[0], TagFields[1],
			      TagFields[2], TagFields[3],
			      TagFields[4], TagFields[5],
			      TagFields[6], TagFields[7],
			      TagFields[8], TagFields[9],
			      TagFields[10], TagFields[11],
			      TagFields[12], TagFields[13],
			      TagFields[14], TagFields[15],
			      TagFields[16], TagFields[17],
			      TagFields[18], TagFields[19],
			      TagFields[20], TagFields[21],
			      TagFields[22], TagFields[23],
			      TagFields[24], TagFields[25],
			      TagFields[26], TagFields[27],
			      TagFields[28], TagFields[29],
			      TagFields[30], TagFields[31]);
      NewXml->Type = (char *) malloc (strlen (ActualTag) + 1);
      if (NewXml->Type == NULL)
	{
	  ReturnValue = 4;
	  goto Error;
	}
      strcpy (NewXml->Type, ActualTag);
      TagToName (ActualTag);
      NewXml->Name = (char *) malloc (strlen (ActualTag) + 1);
      if (NewXml->Name == NULL)
	{
	  ReturnValue = 5;
	  goto Error;
	}
      strcpy (NewXml->Name, ActualTag);
      //  Analyze the tag's variables.
      TagFieldCount--;
      Hinted = 0;
      for (i = 0; i < TagFieldCount; i++)
	{
	  // Clean up the data portion.
	  for (s = TagFields[i]; *s && *s != '='; s++);
	  if (*s != '=' || s[1] != '\"')
	    continue;
	  s += 2;		// s points to start of the data.  
	  for (ss = s; *ss && *ss != '\"'; ss++);
	  if (*ss != '\"')
	    continue;
	  *ss = '\0';		// data area now nul-terminated.
	  // Now that the dat's swellerific, determine what 
	  // type of data it actually is!  In all cases except
	  // TYPE, the data is converted to the "non-tag" form.
	  if (0 == strncasecmp (TagFields[i], "TYPE=", 5))
	    {
	      free (NewXml->Type);	// Remove the default.
	      NewXml->Type = (char *) malloc (strlen (s) + 1);
	      if (NewXml->Type == NULL)
		{
		  ReturnValue = 7;
		  goto Error;
		}
	      strcpy (NewXml->Type, s);
	      continue;
	    }
	  TagToName (s);	// Clean up the data.  
	  if (0 == strncasecmp (TagFields[i], "HINT=", 5))
	    {
	      NewXml->Hint = (char *) malloc (strlen (s) + 1);
	      if (NewXml->Hint == NULL)
		{
		  ReturnValue = 6;
		  goto Error;
		}
	      strcpy (NewXml->Hint, s);
	      Hinted = 1;
	    }
	  else if (0 == strncasecmp (TagFields[i], "A1=", 3))
	    NewXml->Applicability[DO178A_LEVEL_1] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "A2=", 3))
	    NewXml->Applicability[DO178A_LEVEL_2] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "A3=", 3))
	    NewXml->Applicability[DO178A_LEVEL_3] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BA=", 3))
	    NewXml->Applicability[DO178B_LEVEL_A] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BB=", 3))
	    NewXml->Applicability[DO178B_LEVEL_B] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BC=", 3))
	    NewXml->Applicability[DO178B_LEVEL_C] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BD=", 3))
	    NewXml->Applicability[DO178B_LEVEL_D] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "BE=", 3))
	    NewXml->Applicability[DO178B_LEVEL_E] = InterpretLevel (s);
	  else if (0 == strncasecmp (TagFields[i], "PSAC=", 5))
	    NewXml->Documents[PSAC] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SDP=", 4))
	    NewXml->Documents[SDP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SVP=", 4))
	    NewXml->Documents[SVP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SCMP=", 5))
	    NewXml->Documents[SCMP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SQAP=", 5))
	    NewXml->Documents[SQAP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SRS=", 4))
	    NewXml->Documents[SRS] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SDS=", 4))
	    NewXml->Documents[SDS] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SCS=", 4))
	    NewXml->Documents[SCS] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SRD=", 4))
	    NewXml->Documents[SRD] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "DD=", 3))
	    NewXml->Documents[DD] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SVCP=", 5))
	    NewXml->Documents[SVCP] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SECI=", 5))
	    NewXml->Documents[SECI] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SCI=", 4))
	    NewXml->Documents[SCI] = InterpretDoc (s);
	  else if (0 == strncasecmp (TagFields[i], "SAS=", 4))
	    NewXml->Documents[SAS] = InterpretDoc (s);
	}
      if (!Hinted)
	{
	  NewXml->Hint = (char *) malloc (strlen ("") + 1);
	  if (NewXml->Hint == NULL)
	    {
	      ReturnValue = 8;
	      goto Error;
	    }
	  strcpy (NewXml->Hint, "");
	}
      // Get the text, if any.  Note that on the first invocation
      // after program start, Text == NULL and TextFieldSize == 0.
      // The loop is intended to iterate only once, and iterates
      // more than that only if it discovers that the text buffer
      // is too small to hold the data, and hence needs to be
      // enlarged.
      LowWhere = ftell (fp);
      while (1)
	{
	  // Allocate the text buffer, aborting if out of memory.
	  // If reallocating, make the buffer 32K larger than
	  // the last time it was allocated.  In most cases, this
	  // is done only once each time the program runs, because
	  // the Text buffer is persistent.  However, it may be
	  // executed more times if Text has to be enlarged, in
	  // which case the enlarged buffer persists.
	  if (Text == NULL)
	    {
	      TextFieldSize += 32768;
	      printf ("Enlarging text buffer to %ld bytes @%ld.\n",
		      TextFieldSize, LowWhere);
	      if (TextFieldSize == 0x10000)
		{
		  printf
		    ("-------------------------------------------------\n");
		  printf
		    ("NOTE:  If the text fields are TOO large, you will\n");
		  printf
		    ("eventually exceed the capacity of LaTeX (used by\n");
		  printf ("DocBook) to process the documents output by\n");
		  printf
		    ("Do178Builder.  If this happens, it may be fixable\n");
		  printf
		    ("by enlarging the \'hash_extra\' field in the TeX\n");
		  printf
		    ("configuration file texmf.cnf, and then executing\n");
		  printf
		    ("\"fmtutil --all\".  However, this is a TeX issue\n");
		  printf
		    ("rather than a Do178Builder-specific issue, and \n");
		  printf ("you may need to research it as such.\n");
		  printf
		    ("-------------------------------------------------\n");
		}
	      Text = (char *) malloc (TextFieldSize);
	      if (Text == NULL)
		{
		  printf ("Out of memory.\n");
		  LowLevelError = LowWhere;
		  LowReturnCode = 10;
		  ReturnValue = 10;
		  goto Error;
		}
	    }
	  // Fetch the text into the buffer.  If it won't fit,
	  // then deallocate the buffer so that it can be 
	  // enlarged, and loop.
	  switch (GetBetweenTag (fp, Text, TextFieldSize))
	    {
	    case 0:		// success
	      break;
	    case 1:		// buffer too small
	      free (Text);
	      Text = NULL;
	      fseek (fp, LowWhere, SEEK_SET);
	      continue;		// iterate the 'while' again.
	    default:		// file-read error.
	      LowLevelError = LowWhere;
	      LowReturnCode = 9;
	      ReturnValue = 9;
	      goto Error;
	    }
	  // Fixup the text, and save it in the document's
	  // data structures.
	  CleanupText (Text);
	  NewXml->Text = (char *) malloc (strlen (Text) + 1);
	  if (NewXml->Text == NULL)
	    {
	      printf ("Out of memory.\n");
	      LowLevelError = LowWhere;
	      LowReturnCode = 10;
	      ReturnValue = 10;
	      goto Error;
	    }
	  strcpy (NewXml->Text, Text);
	  break;
	}
      // This must be followed either by the closing tag, or else by a
      // tag for a child-tree.  
      Position = ftell (fp);
      if (GetTag (fp, Tag, sizeof (Tag)))
	{
	  ReturnValue = 11;
	  goto Error;
	}
      // If it's a closing tag, and if it matches the opening tag,
      // then we're swell.
      if (Tag[0] == '/')
	{
	  TagToName (&Tag[1]);
	  if (0 == strcmp (&Tag[1], NewXml->Name))
	    continue;
	  else
	    {
	      ReturnValue = 12;
	      goto Error;
	    }
	}

      // If it's not a closing tag, then it must be an opening tag, and
      // we must do a child tree.
      fseek (fp, Position, SEEK_SET);
      if (LoadXmlLevel (Level, fp, NewXml))
	{
	  ReturnValue = 13;
	  goto Error;
	}
    }

  // All done!
Fine:
  ReturnValue = 0;
Error:
  if (ReturnValue != 0 && LowLevelError == -1L)
    {
      LowReturnCode = ReturnValue;
      LowLevelError = Where;
    }
  Level--;
  return (ReturnValue);
}

//------------------------------------------------------------------
// Loads a Do178Builder XML file.
// Returns 0 on success.

int
LoadXmlFile (const char *Filename, struct XmlStruct *Xml)
{
  FILE *fp;
  int ReturnValue = 100;
  char *s;
  //int i;

  fp = fopen (Filename, "rb");
  if (fp == NULL)
    {
      ReturnValue = 101;
      goto Done;
    }
  ClearXml (Xml);

  // The recursive function that reads in the XML isn't going
  // to work unless the in-memory version already has a root
  // element, and the root has been removed from the file.
  // So let's do that.
  if (GetTag (fp, Tag, sizeof (Tag)))
    {
      ReturnValue = 102;
      goto Done;
    }
  for (s = Tag; *s && !isspace (*s); s++);
  *s = '\0';
  Xml->Color = (wxColour *) wxBLACK;
  Xml->Type = (char *) malloc ((s - Tag) + 1);
  if (NULL == Xml->Type)
    {
      ReturnValue = 103;
      goto Done;
    }
  strcpy (Xml->Type, Tag);
  TagToName (Tag);
  Xml->Name = (char *) malloc (strlen (Tag) + 1);
  if (NULL == Xml->Name)
    {
      ReturnValue = 104;
      goto Done;
    }
  strcpy (Xml->Name, Tag);
  Xml->Text = (char *) malloc (1);
  if (NULL == Xml->Text)
    {
      ReturnValue = 105;
      goto Done;
    }
  strcpy (Xml->Text, "");
  Xml->Hint = (char *) malloc (1);
  if (NULL == Xml->Hint)
    {
      ReturnValue = 106;
      goto Done;
    }
  strcpy (Xml->Hint, "");

  // Now actually read the file.
  if (0 != (ReturnValue = LoadXmlLevel (0, fp, Xml)))
    goto Done;

  // This should have left the closing tag for the root level in the
  // buffer.  Let's check!
  if (Tag[0] != '/' || 0 != strcmp (&Tag[1], Xml->Type))
    {
      ReturnValue = 107;
      goto Done;
    }

  // All done!    
  ReturnValue = 0;
  NormalizeApplicability (Xml);
Done:
  if (fp != NULL)
    fclose (fp);
  if (ReturnValue != 0)
    ClearXml (Xml);
  return (ReturnValue);
}

//------------------------------------------------------------------
// Saves a single level of an XML structure to disk.  On entry,
// Xml points to the first element in the level.

// Names of the document types, and other variable names.
// These are just the various constants from Do178Builder.h,
// presented as string arrays.
static const char *DocVarNames[] = {
  "PSAC", "SDP", "SVP", "SCMP", "SQAP", "SRS", "SDS", "SCS",
  "SRD", "DD", "SVCP", "SECI", "SCI", "SAS"
};
static const char *LevelNames[] = {
  "A1", "A2", "A3", "BA", "BB", "BC", "BD", "BE"
};

static int
SaveXmlLevel (int Level, FILE * fp, struct XmlStruct *Xml)
{
  int RetVal = 1;
  int i;
  char *s;
  for (; Xml != NULL; Xml = Xml->Next)
    {
      // Indent.
      for (i = 0; i < Level; i++)
	putc (' ', fp);

      // Write the tag. 
      NameToTag (Xml->Name, Tag);
      fprintf (fp, "<%s", Tag);
      if (0 != strcmp (Xml->Type, Tag))
	{
	  fprintf (fp, " TYPE=\"%s\"", Xml->Type);
	}
      NameToTag (Xml->Hint, Tag);
      fprintf (fp, " HINT=\"%s\"", Tag);
      for (i = 0; i < NUM_DOCUMENTS; i++)
	if (Xml->Documents[i] != 0)
	  fprintf (fp, " %s=\"YES\"", DocVarNames[i]);
      for (i = 0; i < NUM_CRITICALITY_LEVELS; i++)
	if (Xml->Applicability[i] != 0)
	  {
	    fprintf (fp, " %s=\"", LevelNames[i]);
	    switch (Xml->Applicability[i] & SATISFIED_MASK)
	      {
	      case SATISFIED_DISCRETION:
		putc ('N', fp);
		break;
	      case SATISFIED_YES:
		putc ('Y', fp);
		break;
	      case SATISFIED_INDEPENDENCE:
		putc ('I', fp);
		break;
	      default:
		putc ('N', fp);
		break;
	      }
	    switch (Xml->Applicability[i] & CC_MASK)
	      {
	      case CC_DISCRETION:
		putc ('N', fp);
		break;
	      case CC_1:
		putc ('1', fp);
		break;
	      case CC_2:
		putc ('2', fp);
		break;
	      default:
		putc ('N', fp);
		break;
	      }
	    putc ('\"', fp);
	  }
      fprintf (fp, ">");

      // Write the text.
      for (s = Xml->Text; *s; s++)
	{
	  if (*s == '<')
	    fprintf (fp, "&lt;" /*"&#60;" */ );
	  else if (*s == '>')
	    fprintf (fp, "&gt;" /*"&#62;" */ );
	  else if (*s == '&')
	    fprintf (fp, "&amp;" /*"&#38;" */ );
	  else
	    putc (*s, fp);
	}

      // Write the child level, if any.
      if (Xml->Children != NULL)
	{
	  putc ('\n', fp);
	  if (SaveXmlLevel (Level + 1, fp, Xml->Children))
	    goto Done;
	}

      // Write the closing tag.
      NameToTag (Xml->Name, Tag);
      if (Xml->Children != NULL)
	for (i = 0; i < Level; i++)
	  putc (' ', fp);
      fprintf (fp, "</%s>\n", Tag);
    }

  RetVal = 0;
Done:
  return (RetVal);
}

//-----------------------------------------------------------------
// Check to see if a file exists.

int
FileExists (const char *Filename)
{
  FILE *fp;
  fp = fopen (Filename, "r");
  if (fp == NULL)
    return (0);
  fclose (fp);
  return (1);
}

//------------------------------------------------------------------
// Saves a Do178Builder XML file.
// Returns 0 on success.

int
SaveXmlFile (const char *Filename, struct XmlStruct *Xml)
{
  FILE *fp;
  int ReturnValue = 1;
  wxString TempFile, BackFile, DoubleBackFile;
  static char sTF[SOD_SIZE], sBF[SOD_SIZE], sDBF[SOD_SIZE];

  //fp = fopen (Filename, "wb");
  TempFile = SFIX (Filename);
  TempFile = TempFile + wxT (".tmp");
  BackFile = SFIX (Filename);
  BackFile = BackFile + wxT (".bak");
  DoubleBackFile = SFIX (Filename);
  DoubleBackFile = DoubleBackFile + wxT (".bk2");
  strcpy (sTF, TempFile.char_str ());
  strcpy (sBF, BackFile.char_str ());
  strcpy (sDBF, DoubleBackFile.char_str ());
  fp = fopen (sTF, "wb");
  if (fp == NULL)
    goto Done;

  // We have a recursive function to save the file, 
  // because that allows us to descend to lower levels easily.
  if (SaveXmlLevel (0, fp, Xml))
    goto Done;

  // The file has now been created, as a temporary file.
  // We need to move the original to the backup, and the
  // temporary to the original. If it doesn't work, we
  // do our darndest to leave the original file and the
  // primary backup unchanged.  The secondary backup will
  // usually have been deleted.
  if (fclose (fp))
    goto Done;
  fp = NULL;
  
  remove (sDBF);
  if (FileExists (sBF))
    {
      if (rename (sBF, sDBF))
	{
	  goto Done;
	}
    }
  if (FileExists (Filename))
    {
      if (rename (Filename, sBF))
	{
	  rename (sDBF, sBF);
	  goto Done;
	}
    }
  if (rename (sTF, Filename))
    {
      rename (sBF, Filename);
      rename (sDBF, sBF);
      goto Done;
    }

  // All done!    
  ReturnValue = 0;
Done:
  if (fp != NULL)
    fclose (fp);
  return (ReturnValue);
}

//------------------------------------------------------------------------
// Navigates through the XML linked list, finding items by numerical index
// rather than hierarchy.   

struct XmlStruct *
FindXml (struct XmlStruct *Root, wxTreeItemId Index)
{
  XmlStruct *Next, *RetVal;
  for (Next = Root; Next != NULL; Next = Next->Next)
    {
      if (Index == Next->Id)
	return (Next);
      if (Next->Children != NULL)
	{
	  RetVal = FindXml (Next->Children, Index);
	  if (RetVal != NULL)
	    return (RetVal);
	}
    }

  return (NULL);
}

//-------------------------------------------------------------------------
// Determine all of the section numbers.  Can operate recursively.  
// Basically, finds the section number of the closest ancestor relevant to
// the same document, and adds stuff onto that section number.
// On entry, Root is the root of a sub-tree, and Root itself has already 
// been processed, except when Root is itself the root of the entire tree.
//
// In assigning new numbers at a given level, we need to know what numbers
// have already been used at that level.  That's harder to do than it 
// might seem, because the outline structure of the printed documents can
// differ from that of the online data.  The quick&dirty solution is to 
// just keep an array of the used section numbers, and if we find that a 
// number we want to use is taken, we just bump it upward.

static wxStringList *UsedSections[NUM_DOCUMENTS];

void
HtmlSections (struct XmlStruct *Root)
{
  int i, SubSection[NUM_DOCUMENTS];
  struct XmlStruct *Parent, *Record;
  for (i = 0; i < NUM_DOCUMENTS; SubSection[i++] = 1);
  // Loop on the sub-items.
  for (Record = Root->Children; Record != NULL; Record = Record->Next)
    {
      // By default, clear all of the section numbers for this record.
      for (i = 0; i < NUM_DOCUMENTS; i++)
	{
	  *Record->Sections[i] = '\0';
	  Record->SectionLevel[i] = 0;
	}
      // If this record is still named TBDsomething, ignore it!
#ifdef AUTO_ELIMINATE_TBD
      if (0 == strncmp (Record->Name, "TBD", 3))
	continue;
#endif
      // For each document type, search the ancestors to find those
      // relevant to that document type.  The root entry and the 
      // Project_Overall_Information sub-item of the root don't count.
      // Also, any section title "Figure" is really a figure.
      if (0 == strcmp (Record->Type, "Project_Overall_Information"))
	continue;
      if (0 == strcmp (Record->Type, "Figure"))
	continue;
      for (i = 0; i < NUM_DOCUMENTS; i++)
	if (Record->Documents[i] != 0)
	  for (Parent = Record->Parent; Parent != NULL;
	       Parent = Parent->Parent)
	    if (Parent->Parent == NULL)
	      {
		Record->SectionLevel[i] = 0;
		do
		  sprintf (Record->Sections[i], "%d", SubSection[i]++);
		while (UsedSections[i]->Member (SFIX (Record->Sections[i])));
		UsedSections[i]->Add (SFIX (Record->Sections[i]));
		//Record->Sections[i] << SubSection[i]++;
	      }
	    else if (Parent->Documents[i] != 0)
	      {
		Record->SectionLevel[i] = Parent->SectionLevel[i] + 1;
		do
		  sprintf (Record->Sections[i], "%s.%d", Parent->Sections[i],
			   SubSection[i]++);
		while (UsedSections[i]->Member (SFIX (Record->Sections[i])));
		UsedSections[i]->Add (SFIX (Record->Sections[i]));
		//Record->Sections[i] << Parent->Sections[i] << "." << SubSection[i]++;
		break;
	      }
      // Take care of the subtree.
      HtmlSections (Record);
    }
}

//---------------------------------------------------------------------
// Writes an arbitrary text string as paragraph data.  In other words,
// as <para>%s</para><para>%s</para>... where each paragraph is initially
// delimited by newlines.  Returns the number of paragraphs written, 
// or zero on error.

// This array holds all of the inline DocBook tags we're willing to
// recognize.  (Actually, this list contains all of the inline elements
// that I know about from DocBook 3.1.)
const char *InlineTags[] = {
  //"abbref", 
  //"accel",
  //"acronym", 
  //"action",
  //"anchor",   // Requires an attribute.
  //"application",
  "citation",
  //"citerefentry",
  "citetitle",
  "classname",
  "command",
  "computeroutput",
  "constant",
  "database",
  //"email",    // Seems not to be includable in <para>
  "emphasis",
  "envar",
  "errorcode",
  "errorname",
  "errortype",
  "filename",
  "firstterm",
  //"footnote", // Must itself contain <para>
  //"foreignphrase", // Doesn't work, don't know why.
  "function",
  //"glossterm",// "There is no glossary"
  //"guibutton",
  //"guiicon",
  //"guilabel",
  //"guimenu",
  //"guimenuitem",
  //"guisubmenu",
  "hardware",
  //"inlineequation",
  //"inlinegraphic",
  //"interface",
  //"interfacedefinition",
  "keycap",
  "keycode",
  //"keycombo",
  //"keysym",
  //"link",
  //"literal",
  //"markup",
  "medialabel",
  //"menuchoice",
  "mousebutton",
  //"msgtext",
  //"oLink",
  "option",
  "optional",
  "parameter",
  //"phrase",
  "programlisting",
  "prompt",
  "property",
  //"quote",
  "replaceable",
  "returnvalue",
  //"sgmltag",
  //"shortcut",
  "structfield",
  "structname",
  "subscript",
  "superscript",
  "symbol",
  //"systemitem",
  "token",
  "trademark",
  "type",
  //"ulink",
  "userinput",
  "varname"
    //"wordasword",
    //"xref"
};

#define NUM_INLINETAGS (sizeof(InlineTags)/sizeof(const char *))
int NumAllInlines = NUM_INLINETAGS;
static int InlineLengths[NUM_INLINETAGS];
static int InlineLengthsPrepared = 0;
// This array holds just the "common" inlines suitable for novices.
const char *NoviceTags[] = {
  "citetitle",
  "command",
  "emphasis",
  "function",
  "keycap",
  "option",
  "optional",
  "programlisting",
  "subscript",
  "superscript",
  "trademark"
};

#define NUM_NOVICETAGS (sizeof(NoviceTags)/sizeof(const char *))
int NumNovices = NUM_NOVICETAGS;

// Parse variable-length UTF-8 strings, turning them into multi-byte
// character codes, and advancing the into the string data.  (*s) is 
// the pointer into the string data.  (I.e., **s is a character.)
static int
Utf8 (char **s)
{
  unsigned char c0, c1, c2, c3;
  int i;

  c0 = (unsigned char) (*s)[0];
  if (c0 <= 128)
    {
      (*s)++;
      return (c0);
    }
  c1 = (unsigned char) (*s)[1];
  if ((c0 & 0xe0) == 0xc0 && (c1 & 0xc0) == 0x80)
    {
      (*s) += 2;
      i = ((c0 & 0x1f) << 6) | (c1 & 0x3F);
      return (i);
    }
  c2 = (unsigned char) (*s)[2];
  if ((c0 & 0xf0) == 0xe0 && (c1 & 0xc0) == 0x80 && (c2 & 0xc0) == 0x80)
    {
      (*s) += 3;
      i = ((c0 & 0x0F) << 12) | ((c1 & 0x3F) << 6) | (c2 & 0x3F);
      return (i);
    }
  c3 = (unsigned char) (*s)[3];
  if ((c0 & 0xf8) == 0xf0 && (c1 & 0xc0) == 0x80 && (c2 & 0xc0) == 0x80
      && (c3 & 0xc0) == 0x80)
    {
      (*s) += 4;
      i =
	((c0 & 0x07) << 18) | ((c1 & 0x3F) << 12) | ((c2 & 0x3F) << 6) | (c3 &
									  0x3F);
      return (i);
    }
  // Don't know what this is.
  (*s)++;
  return (c0);
}

// Output the next character, translating it to an entity if it's
// UTF-8.  Return the number of bytes read from the starting
// text position.
static int
UniChar (FILE * fp, char *s)
{
  char *ss;
  int i, CharCode;
  ss = s;
  CharCode = Utf8 (&ss);
  if (CharCode < 0x80)
    putc (CharCode, fp);
  else
    {
      // Search for the name of the entity.  Yes, this isn't a very efficient
      // search, and I could pre-sort the list if it proves to be a problem.
      // The names themselves come from entities.h, which I copied from the 
      // /usr/include directory on one of my Linux boxes.
      for (i = 0; i < NR_OF_ENTITIES; i++)
	if (CharCode == entities[i].value)
	  break;
      // If the entity isn't found in the master list, a numerical entity
      // reference is output instead.  These numerical references are perfectly
      // fine in a theoretical sense, but some DocBook parsers gag on them, which
      // is why I go to the trouble of getting the symbolic names.
      if (i < NR_OF_ENTITIES)
	fprintf (fp, "&%s;", entities[i].name);
      else
	fprintf (fp, "&#%d;", CharCode);
    }
  return (ss - s);
}


static int
WriteParagraphs (FILE * fp, int IndentLevel, char *Text,
		 const char *OutputDirectory)
{
  int Paragraphs = 0;
  int ParagraphOpen = 0;
  int InProgramListing = 0;
  int AfterNewline = 0, IsSpace;
  char *s, *ss, *TagName;
  unsigned i;
  // Make sure that all of the InlineLengths are known.
  if (!InlineLengthsPrepared)
    {
      InlineLengthsPrepared = 1;
      for (i = 0; i < NUM_INLINETAGS; i++)
	InlineLengths[i] = strlen (InlineTags[i]);
    }
  // Start writing stuff!  
  if (*Text != '\0')
    {
      for (s = Text; *s;)
	{
	  if (*s == '\r')
	    {
	      s++;
	      continue;
	    }
	  // Ignore everything in a <trace> tag.
	  if (!strncasecmp (s, "<trace>", 7))
	    {
	      for (s += 7; *s && strncasecmp (s, "</trace>", 8); s++);
	      if (*s != 0)
		s += 8;
	      continue;
	    }
	  // If we find a <DocBook> tag, then we start passing
	  // everything through until finding a </DocBook> tag.
	  // When Do178Builder was DocBook 3.1 SGML only, this was
	  // a simple matter, because no translation was needed.
	  // However, if we are outputting DocBook 4.5 XML, we need
	  // to account for the possibility that the commands within
	  // <docbook>...</docbook> are really DocBook 3.1 and need
	  // to be converted to DocBook 4.5 XML.  Of course, we could
	  // just force the user to do this, but as a courtesy we
	  // try to do it ourselves.  The following conversions are
	  // attempted:
	  //    Element names are converted to all lower case.
	  //    Attribute values are quoted.
	  //    <colspec ...> is converted to <colspec ...></colspec>.
	  // The first two are valid in 3.1 SGML as well, so we just
	  // do them all the time.  The </colspec> thing isn't 
	  // legal in 3.1 SGML, so we only do it for 4.5 XML.
	  if (!strncasecmp (s, "<DocBook>", 9))
	    {
	      int InEntry = 0;
	      s += 9;
	      while (*s != '\0')
		{
		  if (!strncasecmp (s, "</DocBook>", 10))
		    {
		      s += 10;
		      break;
		    }
		  else
		    {
		      // Is this a tag?
		      if (*s == '<')
			{
			  char *ss, *sss, c;
			  int IsColspec, LineFeed = 0, IsImagedata;

			  // Is it a comment?
			  if (s[1] == '!' && s[2] == '-' && s[3] == '-')
			    {
			      // It is a comment.  Look for the end of it.
			      ss = strstr (s + 4, "-->");
			      if (ss == NULL)
				{
				  printf
				    ("-----------------------------------------------------\n");
				  printf
				    ("Unterminated SGML/XML comment found.\n");
				  printf
				    ("-----------------------------------------------------\n");
				  putc ('<', fp);
				  s++;
				  continue;
				}
			      else
				{
				  // Advance pointer to the end of the comment.
				  s = ss + 3;
				  continue;
				}
			    }
			  // You know what?  We'll also pass markup like 
			  // <? ... ?> through unchanged.
			  if (s[1] == '?')
			    {
			      // Look for the end of it.
			      ss = strstr (s + 4, "?>");
			      if (ss == NULL)
				{
				  printf
				    ("-----------------------------------------------------\n");
				  printf
				    ("Unterminated <? ... ?> markup found.\n");
				  printf
				    ("-----------------------------------------------------\n");
				  putc ('<', fp);
				  s++;
				  continue;
				}
			      else
				{
				  // Advance pointer to the end of the markup.
				  for (; s < ss + 2; s++)
				    putc (*s, fp);
				  continue;
				}
			    }

			  // Try to find the end of the tag.  I don't 
			  // want to write a full-blown parser here, so
			  // you can fool me!  Mark the end of the 
			  // element name (with sss) when we find it.
			  sss = NULL;
			  for (ss = s + 1; *ss && *ss != '>'; ss++)
			    {
			      if ((*ss == '\n' || *ss == '\r' || *ss == '<')
				  && !LineFeed)
				{
				  char c;
				  c = *ss;
				  *ss = 0;
				  printf
				    ("-----------------------------------------------------\n");
				  printf
				    ("Possible unterminated tag found: \"%s\".\n",
				     s);
				  printf
				    ("To isolate the problem, try the following command:\n");
				  printf
				    ("\tgrep --context=3 \"%s[^>]\" *.sgml\n",
				     s);
				  printf
				    ("-----------------------------------------------------\n");
				  *ss = c;
				  LineFeed = 1;
				}
			      if (isspace (*ss) && sss == NULL)
				sss = ss;
			    }
			  if (sss == NULL)
			    sss = ss;

			  // Does it happen to be a "colspec" or "imagedata" element?
			  c = *sss;
			  *sss = 0;
			  TagName = s + 1;
			  IsColspec = !strcasecmp (TagName, "colspec");
			  IsImagedata = !strcasecmp (TagName, "imagedata");
			  if (!strcasecmp (TagName, "entry"))
			    InEntry++;
			  else if (!strcasecmp (TagName, "/entry"))
			    {
			      if (InEntry)
				InEntry--;
			    }
			  *sss = c;


			  // We have to skip the following if there was an embedded
			  // newline, since the search for attributes will fail.
			  if (*ss == '>' /*&& !LineFeed */ )	// Found end of tag!
			    {
			      // Output start of tag, including leading '<'
			      // and element name converted to lower case.
			      for (; s < sss; s++)
				{
				  // Detect the special case of a tag name that ends with
				  // '/'.
				  if (!x4 && s == sss - 1 && *s == '/')
				    {
				      *s = 0;
				      fprintf (fp, "></%s", TagName);
				      *s = '/';
				      s = sss;
				      break;
				    }
				  else
				    fputc (tolower (*s), fp);
				}

			      // Now loop on the attributes, so that we can
			      // put them in quotes.
			      while (1)
				{
				  // Find the attribute.  Recall that s points
				  // to the first character following the 
				  // name of the element or the previous 
				  // attribute value.  We have to skip 
				  // past any white space.
				  for (; s < ss && isspace (*s); s++);

				  // What if instead of an attribute we've found "/>"?
				  // For SGML output, we need to handle this.  In some
				  // cases like <para/>, the conversion needs to be
				  // <para/> --> <para></para>.  In others like <colspec/>
				  // it needs to be <colspec/> --> <colspec>.
				  if (*s == '/')
				    {
				      if (!x4 && !IsColspec && !IsImagedata)
					fprintf (fp, "></%s", TagName);
				      break;
				    }

				  if (s >= ss)	// No attribute found.
				    break;

				  // Check for a fileref as a special case, because
				  // Do178Builder needs to fetch the graphics file
				  // and save it in the output directory.
				  if (!strncasecmp (s, "fileref=\"", 9))
				    {
				      char *ss;
				      for (ss = s + 9; *ss && *ss != '\"';
					   ss++);
				      if (*ss == '\"')
					{
					  *ss = 0;
					  FetchGraphic (s + 9,
							OutputDirectory);
					  *ss = '\"';
					}
				    }

				  // Find the end of the attribute name.
				  for (sss = s + 1;
				       sss < ss && !isspace (*sss)
				       && *sss != '='; sss++);

				  // Print the attribute name, with the trailing '=' if any.
				  c = sss[1];
				  sss[1] = 0;
				  fprintf (fp, " %s", s);
				  sss[1] = c;

				  if (*sss != '=')	// Attribute has no value.
				    {
				      s = sss;
				      continue;
				    }
				  sss++;
				  if (*sss == '\'' || *sss == '\"')	// Value already quoted.
				    {
				      // Just print out the value, as-is.
				      fputc (*sss, fp);
				      for (s = sss + 1; s < ss && *s != *sss;
					   s++)
					fputc (*s, fp);
				      if (*s == *sss)
					fputc (*s++, fp);
				      // s now points to the first character after the end of
				      // the attribute value.
				      continue;
				    }

				  // Apparently the value wasn't quoted, so quote it.
				  fputc ('\"', fp);
				  for (s = sss; s < ss && !isspace (*s); s++)
				    fputc (*s, fp);
				  fputc ('\"', fp);

				}

			      // Output end of tag.     
			      fputc ('>', fp);
			      // Adjust for it being a "colspec" tag without
			      // an end tag.  It could have an endtag *either*
			      // by virtue of being terminated with "/>", 
			      // *or* being followed by "</colspec>" or "</>".
			      // Remember that ss points to the '>' at the end
			      // of the tag.
			      if (x4 && IsColspec)
				//if (ss[-1] == '/')
				  if (strncasecmp (ss + 1, "</colspec>", 10))
				    if (strncmp (ss + 1, "</>", 3))
				      fprintf (fp, "</colspec>");
			      // ... and same thing for <imagedata> ...
			      if (x4 && IsImagedata)
				//if (ss[-1] == '/')
				if (strncasecmp (ss + 1, "</imagedata>", 12))
				    if (strncmp (ss + 1, "</>", 3))
				      fprintf (fp, "</imagedata>");
			      // Move on.
			      s = ss + 1;
			      //if (!x4 && IsEntry)
			      //  while (*s == '\n' || *s == '\r')
			      //    s++;
			      continue;
			    }
			  // else, didn't find end of tag, so just process
			  // as a normal character by continuing to the 
			  // stuff below.
			}
		      // Output the character.  If we're within a table entry,
		      // we have to skip all newlines and all space following a
		      // newline to avoid the possibility of "pernicious mixed content" 
		      // in DocBook 3.  (There are other cases that could still 
		      // produce it, but this would seem to be the most common one,
		      // due to cutting-and-pasted nicely indented XML.)  Also, 
		      // there are certain characters that I guess 
		      // must be Unicode that you can get from importing docbook stuff 
		      // from OpenOffice that need to be trapped.
		      IsSpace = isspace (*s);
		      if (IsSpace)
			{
			  if (*s == '\n' || *s == '\r')
			    AfterNewline = 1;
			}
		      else
			AfterNewline = 0;
		      if (!x4 && InEntry && AfterNewline && IsSpace)
			s++;
		      else
			s += UniChar (fp, s);
		    }
		}
	      continue;
	    }
	  // Allow <programlisting> markup to pass through.  
	  // This is treated separately from InlineTags, 
	  // because it can contain newlines.  The only exception
	  // is that references to "&" and "<" (rather than
	  // "&amp;" and "&lt;") are fixed. 
	  if (InProgramListing)
	    {
	      if (!strncasecmp (s, "</programlisting>", 17))
		{
		  InProgramListing = 0;
		  fprintf (fp, "</programlisting>\n");
		  s += 17;
		}
	      else
		{
		  if (*s == '&' && strncmp (s, "&amp;", 5)
		      && strncmp (s, "&lt;", 4))
		    fprintf (fp, "&amp;");
		  else if (*s == '<')
		    fprintf (fp, "&lt;");
		  else
		    putc (*s, fp);
		  s++;
		}
	      continue;
	    }
	  else if (!strncasecmp (s, "<programlisting>", 16))
	    {
	      InProgramListing = 1;
	      fprintf (fp, "<programlisting>");
	      s += 16;
	      continue;
	    }
	  if (*s == '\n')
	    {
	      if (ParagraphOpen)
		{
		  fprintf (fp, "</para>\n");
		  //if (EOF == fputs ("</para>\n", fp))
		  //  return (3);
		  //s++;
		  ParagraphOpen = 0;
		}
	      s++;
	      continue;
	    }
	  else if (!ParagraphOpen)
	    {
	      //for (i = 0; i < IndentLevel; i++)
	      //  if (EOF == putc (' ', fp))
	      //    return (3);
	      //if (EOF == fputs ("<P>", fp))
	      //  return (3);
	      fprintf (fp, "%*s<para>", IndentLevel + 1, "");
	      ParagraphOpen = 1;
	      Paragraphs++;
	    }
	  else if (*s == '<')
	    {
	      // We convert '<' to "&#60;" unless we can recognize
	      // it as inline markup. 
	      ss = s + 1 + (s[1] == '/');
	      for (i = 0; i < NUM_INLINETAGS; i++)
		if (!strncasecmp (ss, InlineTags[i], InlineLengths[i])
		    && ss[InlineLengths[i]] == '>')
		  {
		    ss += InlineLengths[i] + 1;
		    // Originally, the output was just the following
		    // line, but the replacement insures that the 
		    // element is all lower-case, as it needs to be
		    // for XML and is acceptable for SGML.
		    //fwrite (s, ss - s, 1, fp);
		    fputc ('<', fp);
		    if (s[1] == '/')
		      fputc ('/', fp);
		    fwrite (InlineTags[i], InlineLengths[i], 1, fp);
		    fputc ('>', fp);
		    s = ss;
		    break;
		  }
	      if (i < NUM_INLINETAGS)
		continue;
	      fprintf (fp, "&#60;");
	      s++;
	    }
	  else if (*s == '>')
	    {
	      fprintf (fp, "&#62;");
	      s++;
	    }
	  else if (*s == '&')
	    {
	      fprintf (fp, "&#38;");
	      s++;
	    }
	  else
	    s += UniChar (fp, s);
	}
      if (ParagraphOpen)
	{
	  if (EOF == fputs ("</para>\n", fp))
	    return (0);
	}
    }
  return (Paragraphs);
}

//-------------------------------------------------------------------------
// Output the section data for a specific SGML document.  Works recursively.
// On entry, Root should already have been processed.  We merely process
// all of the descendents.  Returns:
//      0               success
//      3               disk error
//      0x10            could not find source graphic
//      0x11            could not create destination graphic
//      0x12-0xff       other error

/*
This stuff was here from the olden days, when HTML was output 
rather than SGML.
#define MAX_HEADER_LEVELS 16
static const char *HeaderIn[MAX_HEADER_LEVELS] = {
  "<H1>", "<H2>", "<H3>", "<H4>",
  "<P><B>", "<P><B>", "<P><B>", "<P><B>",
  "<P><B>", "<P><B>", "<P><B>", "<P><B>",
  "<P><B>", "<P><B>", "<P><B>", "<P><B>"
};
static const char *HeaderOut[MAX_HEADER_LEVELS] = {
  "</H1>\n", "</H2>\n", "</H3>\n", "</H4>\n",
  "</B></P>\n", "</B></P>\n", "</B></P>\n", "</B></P>\n",
  "</B></P>\n", "</B></P>\n", "</B></P>\n", "</B></P>\n",
  "</B></P>\n", "</B></P>\n", "</B></P>\n", "</B></P>\n"
};
*/

// Here are some macros that make it a little easier to deal with
// combined SRD/SDD documents.
#define DOC_IS_SRD ((DocNum == SRD) || (CombineSrdSdd && DocNum == SDD))
#define DOC_IS_SDD ((DocNum == SDD) || (CombineSrdSdd && DocNum == SRD))
#define NEXT_IS_SRD (Next->Documents[SRD] || CombineSrdSdd && Next->Documents[SDD])
#define NEXT_IS_SDD (Next->Documents[SDD] || CombineSrdSdd && Next->Documents[SRD])
#define DOC_IS_PLAN (DocNum == PSAC || DocNum == SDP || DocNum == SVP || DocNum == SQAP || DocNum == SCMP)
#define NEXT_IS_PLAN (Next->Documents[PSAC] || Next->Documents[SDP] || Next->Documents[SVP] || Next->Documents[SQAP] || Next->Documents[SCMP])
#define NEXT_IS_SAME ( \
  (Next->Documents[DocNum] != 0) || \
  (DOC_IS_SRD && NEXT_IS_SRD) || \
  (DOC_IS_SDD && NEXT_IS_SDD) || \
  (CombinedPlanName != NULL && DOC_IS_PLAN && NEXT_IS_PLAN))
#define DOC_IS_PSAC (DocNum == PSAC || (CombinedPlanName != NULL && DOC_IS_PLAN))

static int
OutputSections (FILE * fp, int IndentLevel, int DocNum,
		struct XmlStruct *Root, const char *OutputDirectory)
{
  extern int Do254;
  struct XmlStruct *Next;
  int RetVal, ParagraphOpen, AnythingWritten, OpenSection;
  //char *s;
  //IndentLevel++;
  for (Next = Root->Children; Next != NULL; Next = Next->Next)
    {
#ifdef AUTO_ELIMINATE_TBD
      if (0 == strncmp (Next->Name, "TBD", 3))
	continue;
#endif
      OpenSection = 0;
      if (NEXT_IS_SAME)
	{
	  // If it's a figure, take care of it and continue.
	  if (0 == strcmp (Next->Type, "Figure"))
	    {
	      struct XmlStruct *Field, *FigField, *CapField;
	      // Find the filename:
	      FigField = NULL;
	      CapField = NULL;
	      for (Field = Next->Children; Field != NULL; Field = Field->Next)
		if (0 == strcmp (Field->Type, "Filename"))
		  FigField = Field;
		else if (0 == strcmp (Field->Type, "Caption"))
		  CapField = Field;
	      if (FigField == NULL)	// Malformed -- no Filename field.
		continue;
	      if (CapField != NULL)
		fprintf (fp, "%*s<figure><title>%s</title>", IndentLevel, "",
			 CapField->Text);
	      FetchGraphic (FigField->Text, OutputDirectory);
	      fprintf (fp,
		       "<graphic fileref=\"%s\" align=\"center\"></graphic>",
		       FigField->Text);
	      if (CapField != NULL)
		fprintf (fp, "</figure>");
	      if (EOF == putc ('\n', fp))
		return (3);
	      continue;
	    }
	  // Write the section header.
	  OpenSection = 1;
	  fprintf (fp, "%*s<section>\n", IndentLevel, "");
	  fprintf (fp, "%*s<title>%s</title>\n", IndentLevel + 1, "",
		   Next->Name);
	  AnythingWritten = 0;
	  ParagraphOpen = 0;
	  // Write traceability data.  Traceability data is determined by the following
	  // method:  Firstly, the Type of the item must begin with TBD.  If the item
	  // is relevant to both SRD & SDD, or SAS & SVCP, or SDD & SVCP, then there is
	  // traceability between them.
	  if (strncmp (Next->Type, "TBD", 3) == 0)
	    {
	      int NoteOpen = 0;
	      const char *Doc1, *Doc2;
	      char *st, *ste;
	      st = strstr (Next->Text, "<trace>");
	      ste = NULL;
	      if (st != NULL)
		ste = strstr (st + 7, "</trace>");
	      if (Next->Documents[SVCP])
		{
		  if (DocNum == SVCP && ste != NULL
		      && !strcmp (Next->Type, "TBD_Test_Case"))
		    {
		      if (!NoTrace)
			{
			  if (!NoForwardTrace)
			    {
			      if (!NoteOpen)
				{
				  NoteOpen = 1;
				  fprintf (fp,
					   "%*s<note><title>Traceability</title>",
					   IndentLevel + 1, "");
				}
			      //fprintf (fp,
			      //       "<para>SDD section %s &rarr; SVCP section %s.</para>",
			      //      Next->Sections[SDD], Next->Sections[SVCP]);
			      *ste = 0;
			      if (Do254)
				Doc1 = ("HTPR");
			      else
				Doc1 = ("SVCP");
			      fprintf (fp,
				       "<para>%s section %s &rarr; %s.</para>",
				       Doc1, Next->Sections[SVCP], st + 7);
			      *ste = '<';
			      //fprintf (fp, "</note>\n");
			    }
			}
		    }
		  if (NEXT_IS_SDD && (DocNum == SVCP || DOC_IS_SDD))
		    {
		      // This is for SDD -> SVCP
		      if (Do254)
			{
			  if (CombineSrdSdd)
			    Doc1 = ("HRD/HDRD");
			  else
			    Doc1 = ("HDRD");
			  Doc2 = ("HTPR");
			}
		      else
			{
			  if (CombineSrdSdd)
			    Doc1 = ("SRD/SDD");
			  else
			    Doc1 = ("SDD");
			  Doc2 = ("SVCP");
			}
		      if (!NoTrace)
			{
			  if (!NoForwardTrace || DocNum == SVCP)
			    {
			      if (!NoteOpen)
				{
				  NoteOpen = 1;
				  fprintf (fp,
					   "%*s<note><title>Traceability</title>",
					   IndentLevel + 1, "");
				}
			      if (ste == NULL || !DOC_IS_SDD || NEXT_IS_SRD)
				{
				  fprintf (fp,
					   "<para>%s section %s &rarr; %s section %s.</para>",
					   Doc1, Next->Sections[SDD],
					   Doc2, Next->Sections[SVCP]);
				}
			      else
				{
				  *ste = 0;
				  fprintf (fp,
					   "<para>%s section %s &rarr; "
					   "%s section %s.</para>",
					   Doc1, Next->Sections[SDD],
					   Doc2, Next->Sections[SVCP]);
				  fprintf (fp,
					   "<para>%s section %s &rarr; "
					   "%s.</para>",
					   Doc1, Next->Sections[SDD], st + 7);
				  *ste = '<';
				}
			    }
			}
		      //fprintf (fp, "</note>\n");
		    }
		  else if (NEXT_IS_SRD && (DocNum == SVCP || DOC_IS_SRD))
		    {
		      // This is for SRD -> SVCP.
		      if (Do254)
			{
			  if (CombineSrdSdd)
			    Doc1 = ("HRD/HDRD");
			  else
			    Doc1 = ("HRD");
			  Doc2 = ("HTPR");
			}
		      else
			{
			  if (CombineSrdSdd)
			    Doc1 = ("SRD/SDD");
			  else
			    Doc1 = ("SRD");
			  Doc2 = ("SVCP");
			}
		      if (!NoTrace)
			{
			  if (!NoForwardTrace || DocNum == SVCP)
			    {
			      if (!NoteOpen)
				{
				  NoteOpen = 1;
				  fprintf (fp,
					   "%*s<note><title>Traceability</title>",
					   IndentLevel + 1, "");
				}
			      fprintf (fp,
				       "<para>%s section %s &rarr; %s section %s.</para>",
				       Doc1, Next->Sections[SRD],
				       Doc2, Next->Sections[SVCP]);
			    }
			}
		      //fprintf (fp, "</note>\n");
		    }
		  else if (Next->Documents[SAS]
			   && (DocNum == SVCP || DocNum == SAS)
			   && !(DocNum == SAS
				&& (NEXT_IS_SRD || NEXT_IS_SDD)))
		    {
		      if (Do254)
			{
			  Doc1 = ("HAS");
			  Doc2 = ("HTPR");
			}
		      else
			{
			  Doc1 = ("SAS");
			  Doc2 = ("SVCP");
			}
		      // This is for test cases arising from changes since prior release.
		      // I.e., SAS -> SVCP
		      if (!NoTrace)
			{
			  if (!NoForwardTrace || DocNum == SAS)
			    {
			      if (!NoteOpen)
				{
				  NoteOpen = 1;
				  fprintf (fp,
					   "%*s<note><title>Traceability</title>",
					   IndentLevel + 1, "");
				}
			      fprintf (fp,
				       "<para>%s section %s &rarr; %s section %s.</para>",
				       Doc1, Next->Sections[SAS],
				       Doc2, Next->Sections[SVCP]);
			    }
			}
		      //fprintf (fp, "</note>\n");
		    }
		}
	      if (NEXT_IS_SRD && NEXT_IS_SDD && (DOC_IS_SRD || DOC_IS_SDD))
		{
		  if (Do254)
		    {
		      Doc1 = ("HRD");
		      Doc2 = ("HDRD");
		    }
		  else
		    {
		      Doc1 = ("SRD");
		      Doc2 = ("SDD");
		    }
		  if (!NoTrace)
		    {
		      // This is for SRD -> SDD
		      if (!NoForwardTrace || DOC_IS_SDD)
			{
			  if (!(CombineSrdSdd && !Do254))
			    {
			      if (!NoteOpen)
				{
				  NoteOpen = 1;
				  fprintf (fp,
					   "%*s<note><title>Traceability</title>",
					   IndentLevel + 1, "");
				}
			      fprintf (fp,
				       "<para>%s section %s &rarr; %s section %s.</para>",
				       Doc1, Next->Sections[SRD],
				       Doc2, Next->Sections[SDD]);
			      //fprintf (fp, "</note>\n");
			    }
			}
		      // This is for SRM -> SRD
		      if (ste != NULL && DOC_IS_SRD)
			{
			  if (!NoteOpen)
			    {
			      NoteOpen = 1;
			      fprintf (fp,
				       "%*s<note><title>Traceability</title>",
				       IndentLevel + 1, "");
			    }
			  *ste = 0;
			  fprintf (fp,
				   "<para>%s &rarr; %s section %s</para>",
				   st + 7, Doc1, Next->Sections[SRD]);
			  *ste = '<';
			}
		    }
		}
	      if (NoteOpen)
		{
		  NoteOpen = 0;
		  fprintf (fp, "</note>\n");
		}
	    }
	  // Write the section data.
	  //if (0 == WriteParagraphs (fp, IndentLevel, Next->Text))
	  //  if (EOF == fprintf (fp, "%*s<para>(TBD)</para>", IndentLevel, ""))
	  //    return (3);
	  WriteParagraphs (fp, IndentLevel, Next->Text, OutputDirectory);
	}
      RetVal =
	OutputSections (fp, IndentLevel + OpenSection, DocNum, Next,
			OutputDirectory);
      if (RetVal != 0)
	return (RetVal);
      if (OpenSection)
	fprintf (fp, "%*s</section>\n", IndentLevel, "");
    }
  return (0);
}

//-------------------------------------------------------------------------
// Generate DocBook SGML for a specific DO-178B document.  The numbering 
// of the docs is the same as the constants PSAC, SDP, SVP, etc.  Returns:
//      0       Success
//      1       Unknown doc type.
//      2       Cannot create file.
//      3       Disk-write error.

static int
MakeSgmlDocument (int DocNum, struct XmlStruct *Root,
		  const char *OutputDirectory)
{
  extern int Do254, NoProductname;
  int ReturnValue;
  FILE *fp = NULL;
  char DocAcronym[129], DocName[64], Filename[1024], XxxxIdentifier[144];
  struct XmlStruct *Record, *Parent;
  char *Company_Name = NULL, *Project_Name = NULL, *Author = NULL,
    *Document_Version = NULL, *Publication_Date = NULL,
    *CopyrightYears = NULL, *CopyrightHolder = NULL,
    *Legal_Notice = NULL, *FigureFilename = NULL, *XXXX_Identifier = NULL;
  const char *DocFilenameAcronym = NULL, *DocIdentification = NULL, *Extension;
  // Determine the document name.
  switch (DocNum)
    {
    case PSAC:
      if (Do254)
	{
	  if (CombinedPlanName != NULL)
	    {
	      if (!strcmp (CombinedPlanName, "PHAC"))
	        {
		  strcpy (DocName, "Plan for Hardware Aspects of Certification");
		  strcpy (DocAcronym, "PHAC");
		}
	      else if (!strcmp (CombinedPlanName, "HDP"))
	        {
		  strcpy (DocName, "Hardware Design Plan");
		  strcpy (DocAcronym, "HDP");
		}
	      else if (!strcmp (CombinedPlanName, "HVVP"))
	        {
		  strcpy (DocName, "Hardware Validation and Verification Plan");
		  strcpy (DocAcronym, "HVVP");
		}
	      else if (!strcmp (CombinedPlanName, "HCMP"))
	        {
		  strcpy (DocName, "Hardware Configuration Management Plan");
		  strcpy (DocAcronym, "HCMP");
		}
	      else if (!strcmp (CombinedPlanName, "HPAP"))
	        {
		  strcpy (DocName, "Hardware Process Assurance Plan");
		  strcpy (DocAcronym, "HPAP");
		}
	    }
	  else
	    {
	      strcpy (DocName, "Plan for Hardware Aspects of Certification");
	      strcpy (DocAcronym, "PHAC");
	    }
	}
      else
	{
	  if (CombinedPlanName != NULL)
	    {
	      if (!strcmp (CombinedPlanName, "PSAC"))
	        {
		  strcpy (DocName, "Plan for Software Aspects of Certification");
		  strcpy (DocAcronym, "PSAC");
		}
	      else if (!strcmp (CombinedPlanName, "SDP"))
	        {
		  strcpy (DocName, "Software Development Plan");
		  strcpy (DocAcronym, "SDP");
		}
	      else if (!strcmp (CombinedPlanName, "SVP"))
	        {
		  strcpy (DocName, "Software Verification Plan");
		  strcpy (DocAcronym, "SVP");
		}
	      else if (!strcmp (CombinedPlanName, "SCMP"))
	        {
		  strcpy (DocName, "Software Configuration Management Plan");
		  strcpy (DocAcronym, "SCMP");
		}
	      else if (!strcmp (CombinedPlanName, "SQAP"))
	        {
		  strcpy (DocName, "Software Quality Assurance Plan");
		  strcpy (DocAcronym, "SQAP");
		}
	    }
	  else
	    {
	      strcpy (DocName, "Plan for Software Aspects of Certification");
	      strcpy (DocAcronym, "PSAC");
	    }
	}
      break;
    case SDP:
      if (CombinedPlanName != NULL)
        return (0);
      if (Do254)
	{
	  strcpy (DocName, "Hardware Design Plan");
	  strcpy (DocAcronym, "HDP");
	}
      else
	{
	  strcpy (DocName, "Software Development Plan");
	  strcpy (DocAcronym, "SDP");
	}
      break;
    case SVP:
      if (CombinedPlanName != NULL)
        return (0);
      if (Do254)
	{
	  strcpy (DocName, "Hardware Validation and Verification Plan");
	  strcpy (DocAcronym, "HVVP");
	}
      else
	{
	  strcpy (DocName, "Software Verification Plan");
	  strcpy (DocAcronym, "SVP");
	}
      break;
    case SCMP:
      if (CombinedPlanName != NULL)
        return (0);
      if (Do254)
	{
	  strcpy (DocName, "Hardware Configuration Management Plan");
	  strcpy (DocAcronym, "HCMP");
	}
      else
	{
	  strcpy (DocName, "Software Configuration Management Plan");
	  strcpy (DocAcronym, "SCMP");
	}
      break;
    case SQAP:
      if (CombinedPlanName != NULL)
        return (0);
      if (Do254)
	{
	  strcpy (DocName, "Hardware Process Assurance Plan");
	  strcpy (DocAcronym, "HPAP");
	}
      else
	{
	  strcpy (DocName, "Software Quality Assurance Plan");
	  strcpy (DocAcronym, "SQAP");
	}
      break;
    case SRS:
      if (Do254)
	{
	  strcpy (DocName, "Requirements Standards");
	  strcpy (DocAcronym, "RS");
	}
      else
	{
	  strcpy (DocName, "Software Requirements Standards");
	  strcpy (DocAcronym, "SRS");
	}
      break;
    case SDS:
      if (Do254)
	{
	  strcpy (DocName, "Hardware Design Standards");
	  strcpy (DocAcronym, "HDS");
	}
      else
	{
	  strcpy (DocName, "Software Design Standards");
	  strcpy (DocAcronym, "SDS");
	}
      break;
    case SCS:
      if (Do254)
	{
	  strcpy (DocName, "Validation and Verification Standards");
	  strcpy (DocAcronym, "VVS");
	}
      else
	{
	  strcpy (DocName, "Software Code Standards");
	  strcpy (DocAcronym, "SCS");
	}
      break;
    case SRD:
      if (Do254)
	{
	  if (CombineSrdSdd)
	    {
	      strcpy (DocName, "Hardware Requirements and Design Data");
	      strcpy (DocAcronym, "HRD/HDRD");
	      DocFilenameAcronym = "HRD-HDRD";
	      DocIdentification = "HRD&#47;HDRD";
	    }
	  else
	    {
	      strcpy (DocName, "Hardware Requirements Data");
	      strcpy (DocAcronym, "HRD");
	    }
	}
      else
	{
	  if (CombineSrdSdd)
	    {
	      strcpy (DocName, "Software Requirements and Design Data");
	      strcpy (DocAcronym, "SRD/SDD");
	      DocFilenameAcronym = "SRD-SDD";
	      DocIdentification = "SRD&#47;SDD";
	    }
	  else
	    {
	      strcpy (DocName, "Software Requirements Data");
	      strcpy (DocAcronym, "SRD");
	    }
	}
      break;
    case DD:
      if (CombineSrdSdd)
	return (0);
      if (Do254)
	{
	  strcpy (DocName, "Hardware Design Representation Data");
	  strcpy (DocAcronym, "HDRD");
	}
      else
	{
	  strcpy (DocName, "Software Design Description");
	  strcpy (DocAcronym, "SDD");
	}
      break;
    case SVCP:
      if (Do254)
	{
	  strcpy (DocName, "Hardware Test Procedures and Results");
	  strcpy (DocAcronym, "HTPR");
	}
      else
	{
	  strcpy (DocName, "Software Verification Cases and Procedures");
	  strcpy (DocAcronym, "SVCP");
	}
      break;
    case SECI:
      if (Do254)
	{
	  strcpy (DocName, "Hardware Acceptance Test Criteria");
	  strcpy (DocAcronym, "HATC");
	}
      else
	{
	  strcpy (DocName,
		  "Software Life Cycle Environment Configuration Index");
	  strcpy (DocAcronym, "SECI");
	}
      break;
    case SCI:
      if (Do254)
	{
	  strcpy (DocName, "Hardware Archive Standards");
	  strcpy (DocAcronym, "HARS");
	}
      else
	{
	  strcpy (DocName, "Software Configuration Index");
	  strcpy (DocAcronym, "SCI");
	}
      break;
    case SAS:
      if (Do254)
	{
	  strcpy (DocName, "Hardware Accomplishment Summary");
	  strcpy (DocAcronym, "HAS");
	}
      else
	{
	  strcpy (DocName, "Software Accomplishment Summary");
	  strcpy (DocAcronym, "SAS");
	}
      break;
    default:
      ReturnValue = 1;
      goto Done;
    }
  if (DocIdentification == NULL)
    DocIdentification = DocAcronym;
  strcpy (XxxxIdentifier, DocIdentification);
  strcat (XxxxIdentifier, "_Identification");
  printf ("XxxxIdentifier \"%s\"\n", XxxxIdentifier);

  // Create the SGML file.
  if (DocFilenameAcronym == NULL)
    DocFilenameAcronym = DocAcronym;
  if (x4)
    Extension = "xml";
  else
    Extension = "sgml";
  sprintf (Filename, "%s%s.%s", OutputDirectory, DocFilenameAcronym,
	   Extension);
  fp = fopen (Filename, "w");
  if (fp == NULL)
    {
      ReturnValue = 2;
      goto Done;
    }

  // Generate the article header.  This would include the title page.
  // [We use DocBook 3.1 (4.1 is actually the current version when this was
  // started) because 3.1 is the one covered in the edition of 
  // "DocBook: The Definitive Guide" that I have.]  
  // The first step is to parse the "Project Overall Information" section to
  // determine what actually goes into the header.
  //printf ("Here %d!\n", DocNum);
  for (Record = Root->Children; Record != NULL; Record = Record->Next)
    if (0 == strcmp (Record->Type, "Project_Overall_Information")
	&& Record->Children != NULL)
      {
	for (Record = Record->Children; Record != NULL; Record = Record->Next)
	  {
	    if (0 == strcmp (Record->Type, "Company_Name") &&
		NULL != Record->Text && 0 != strncmp (Record->Text, "TBD", 3))
	      Company_Name = Record->Text;
	    else if (0 == strcmp (Record->Type, "Project_Name") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Project_Name = Record->Text;
	    else if (0 == strcmp (Record->Type, "Author") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Author = Record->Text;
	    else if (0 == strcmp (Record->Type, "Document_Version"))
	      {
		struct XmlStruct *Where;
		if (NULL != Record->Text
		    && 0 != strncmp (Record->Text, "TBD", 3))
		  Document_Version = Record->Text;
		Where = Record;
		for (Record = Record->Children; Record != NULL;
		     Record = Record->Next)
		  if (0 == strcmp (Record->Type, XxxxIdentifier))
		    {
		      if (NULL != Record->Text
			  && 0 != strncmp (Record->Text, "TBD", 3))
			{
			  XXXX_Identifier = Record->Text;
			  printf ("XXXX_Identifier \"%s\"\n",
				  XXXX_Identifier);
			}
		      break;
		    }
		Record = Where;
	      }
	    else if (0 == strcmp (Record->Type, "Publication_Date") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Publication_Date = Record->Text;
	    else if (0 == strcmp (Record->Type, "Copyright")
		     && Record->Children != NULL)
	      {
		Parent = Record;
		for (Record = Record->Children; Record != NULL;
		     Record = Record->Next)
		  {
		    if (0 == strcmp (Record->Type, "Years") &&
			NULL != Record->Text &&
			0 != strncmp (Record->Text, "TBD", 3))
		      CopyrightYears = Record->Text;
		    else if (0 == strcmp (Record->Type, "Holder") &&
			     NULL != Record->Text &&
			     0 != strncmp (Record->Text, "TBD", 3))
		      CopyrightHolder = Record->Text;
		  }
		Record = Parent;
	      }
	    else if (0 == strcmp (Record->Type, "Legal_Notice") &&
		     NULL != Record->Text &&
		     0 != strncmp (Record->Text, "TBD", 3))
	      Legal_Notice = Record->Text;
	    else if (0 == strcmp (Record->Type, "Figure")
		     && Record->Children != NULL)
	      {
		Parent = Record;
		for (Record = Record->Children; Record != NULL;
		     Record = Record->Next)
		  {
		    if (0 == strcmp (Record->Type, "Filename") &&
			NULL != Record->Text &&
			0 != strncmp (Record->Text, "TBD", 3))
		      FigureFilename = Record->Text;
		  }
		Record = Parent;
	      }
	  }
	break;
      }
  //printf ("Now!\n");
  if (x4)
    {
      fprintf (fp,
	       "<!DOCTYPE article PUBLIC "
	       "\"-//OASIS//DTD DocBook XML V%s//EN\" "
	       "\"http://www.oasis-open.org/docbook/xml/%s/docbookx.dtd\">\n",
	       x4v1, x4v2);
      fprintf (fp, "<article>\n");
      fprintf (fp, " <articleinfo>\n");
    }
  else
    {
      fprintf (fp,
	       "<!DOCTYPE article PUBLIC \"-//OASIS//DTD DocBook V3.1//EN\">\n");
      fprintf (fp, "<article>\n");
      fprintf (fp, " <artheader>\n");
    }
  if (Company_Name != NULL)
    fprintf (fp, "  <corpname>%s</corpname>\n", Company_Name);
  if (Project_Name != NULL)
    {
      if (NoProductname)
	fprintf (fp, "  %s\n", Project_Name);
      else
	fprintf (fp, "  <productname>%s</productname>\n", Project_Name);
    }
  fprintf (fp, "  <title>");
  if (Project_Name != NULL)
    {
      if (NoProductname)
	fprintf (fp, "%s, ", Project_Name);
      else
	fprintf (fp, "<productname>%s</productname>, ", Project_Name);
    }
  fprintf (fp, "%s</title>\n", DocName);
  fprintf (fp, "  <titleabbrev>");
  if (Project_Name != NULL)
    {
      if (NoProductname)
	fprintf (fp, "%s, ", Project_Name);
      else
	fprintf (fp, "<productname>%s</productname>, ", Project_Name);
    }
  fprintf (fp, "%s</titleabbrev>\n", DocAcronym);
  if (Author != NULL)
    fprintf (fp, "  <corpauthor>%s</corpauthor>\n", Author);
  if (Document_Version != NULL && XXXX_Identifier != NULL)
    fprintf (fp, "  <releaseinfo>%s%s</releaseinfo>\n", Document_Version,
	     XXXX_Identifier);
  else if (Document_Version != NULL)
    fprintf (fp, "  <releaseinfo>%s</releaseinfo>\n", Document_Version);
  else if (XXXX_Identifier != NULL)
    fprintf (fp, "  <releaseinfo>%s</releaseinfo>\n", XXXX_Identifier);
  if (Publication_Date != NULL)
    fprintf (fp, "  <pubdate>%s</pubdate>\n", Publication_Date);
  if (CopyrightYears != NULL && CopyrightHolder != NULL)
    fprintf (fp,
	     "  <copyright><year>%s</year><holder>%s</holder></copyright>\n",
	     CopyrightYears, CopyrightHolder);
  if (Legal_Notice != NULL)
    {
      fprintf (fp, "  <legalnotice>\n");
      if (0 == WriteParagraphs (fp, 3, Legal_Notice, OutputDirectory))
	if (EOF == fprintf (fp, "   <para>(TBD)</para>"))
	  return (3);
      fprintf (fp, "  </legalnotice>\n");
    }
  if (FigureFilename != NULL)
    {
      FetchGraphic (FigureFilename, OutputDirectory);
      fprintf (fp, "  <graphic fileref=\"%s\" align=\"center\"></graphic>\n",
	       FigureFilename);
    }
  if (x4)
    fprintf (fp, " </articleinfo>\n");
  else
    fprintf (fp, " </artheader>\n");
  if (fputs ("\n", fp) == EOF)
    {
    DiskError:
      ReturnValue = 3;
      goto Done;
    }

  // Now output the sections.
  OutputSections (fp, 1, DocNum, Root, OutputDirectory);

  // Close the SGML file.
  if (fputs ("\n</article>\n", fp) == EOF)
    goto DiskError;

  ReturnValue = 0;
Done:
  if (fp != NULL)
    fclose (fp);
  return (ReturnValue);
}

//-------------------------------------------------------------------------
// Write the mapping of XML sections to SGML sections to a log file.  
// This is useful in performing diffs across versions. The function acts 
// recursively.

void
HtmlSectionLog (int Indent, FILE * Log, struct XmlStruct *Root)
{
  struct XmlStruct *Record;
  int i;
  uLong Crc;
  for (Record = Root->Children; Record != NULL; Record = Record->Next)
    {
      fprintf (Log, "%d ", Indent);
      for (i = 0; i < NUM_DOCUMENTS; i++)
	if (Record->Sections[i][0] != '\0')
	  fprintf (Log, "%s=%s ", DocVarNames[i], Record->Sections[i]);
      Crc = crc32 (0, (const Bytef *) Record->Text, strlen (Record->Text));
      fprintf (Log, "CRC=0x%08lX NAME=%s\n", Crc, Record->Name);
      if (Record->Children != NULL)
	HtmlSectionLog (Indent + 1, Log, Record);
    }
}

//-------------------------------------------------------------------------
// Output SGML DO-178B documents.  Returns 0 on success.  Otherwise,
// returns the doc number in bits 8-15 and an error code in bits 0-7.

int
XmlToSgml (struct XmlStruct *Root, const char *OutputDirectory)
{
  char *s;
  int i, RetVal = 0;
  // First pass:  Determine all of the section numbers.
  for (i = 0; i < NUM_DOCUMENTS; i++)
    {
      UsedSections[i] = new wxStringList;
      UsedSections[i]->Clear ();
    }
  HtmlSections (Root);
  // Make a log of the document-section to XML mapping.
  s = (char *) malloc (strlen (OutputDirectory) + 13);
  if (s != NULL)
    {
      FILE *Log;
      sprintf (s, "%sSections.xml", OutputDirectory);
      SaveXmlFile (s, Root);
      sprintf (s, "%sSections.log", OutputDirectory);
      Log = fopen (s, "w");
      if (Log != NULL)
	{
	  HtmlSectionLog (1, Log, Root);
	  fclose (Log);
	}
      free (s);
    }
  // Second pass:  Generate SGML for each of the documents.
  for (i = 0; i < NUM_DOCUMENTS; i++)
    {
      RetVal = MakeSgmlDocument (i, Root, OutputDirectory);
      if (RetVal != 0)
	{
	  RetVal |= (i << 8);
	  break;
	}
    }
  for (i = 0; i < NUM_DOCUMENTS; i++)
    delete UsedSections[i];
  return (RetVal);
}

//-------------------------------------------------------------------------
// Prune a tree to a desired assurance level.  We assume the Applicability[]
// data has been normalized (see NormalizeApplicability above), so that 
// the decision as to whether to prune or not can be made completely by 
// examining an item without descending to its children.  Items with 
// Applicability[] identically 0 are not pruned, because they are assumed
// to be items I've added from my own incredible good sense, rather than
// items dictated by DO-178B or DO-254.  We change the color of an item 
// if any of its children have been pruned.  The Root is never pruned, only
// its children.

void
PruneToAssuranceLevel (struct XmlStruct *Root, int AssuranceLevel)
{
  //extern int Do254;
  static wxColour AppChangedColour;
  struct XmlStruct *Next, *Prior = NULL, *NextNext;
  int i;

  if (Root == NULL)
    return;
  if (AssuranceLevel < 0 || AssuranceLevel >= NUM_CRITICALITY_LEVELS)
    return;
  // Loop on the children.
  for (Next = Root->Children; Next != NULL;)
    {
      int NotPrunable;
      //if (!strcmp (Next->Name, "High-Level Description"))
      //  i = 0;
      // Determine if the child is prunable.  A child is NOT prunable
      // if its applicability array is identically 0, or if the specific
      // applicability for this assurance level is non-zero.
      NotPrunable = (Next->Applicability[AssuranceLevel] != 0);
      if (!NotPrunable)
	{
	  for (i = 0; i < NUM_CRITICALITY_LEVELS; i++)
	    if (Next->Applicability[i] != 0)
	      break;
	  if (i >= NUM_CRITICALITY_LEVELS)
	    NotPrunable = 1;
	}
      // If prunable, then prune it.  Otherwise, check its children.
      if (NotPrunable)
	{
	  // Check children level.
	  PruneToAssuranceLevel (Next, AssuranceLevel);
	  Prior = Next;
	  Next = Next->Next;
	}
      else
	{
	  // Prune it!
	  NextNext = Next->Next;
	  Next->Next = NULL;
	  Next->Parent = NULL;
	  ClearXml (Next);
	  free (Next);
	  Next = NextNext;
	  if (Prior == NULL)
	    Root->Children = Next;
	  else
	    Prior->Next = Next;
	  AppChangedColour = wxColour (255, 0, 0);
	  for (NextNext = Root; NextNext != NULL; NextNext = NextNext->Parent)
	    NextNext->Color = &AppChangedColour;
	}
    }
}
